(define-module (sx basic))
(use-modules (sx undo) (sx gui-entry) (sx property))
(export objects-clear-all object-valid object-property-print
        point-valid-basic  point-add-basic  point-del-basic
        link-valid-basic   link-add-basic   link-del-basic
        triang-valid-basic triang-add-basic triang-del-basic
        quad-valid-basic   quad-add-basic   quad-del-basic
        box-valid-basic    box-add-basic    box-del-basic
        group-valid-basic  group-create-basic group-explode-basic
        group-get-points-basic group-get-objects-basic
)

;;; Basic entity infrastructure

; this function removes all objects from the model
(define (objects-clear-all)
  ; reset object database
  (set! *POINT_UNIQUE*  1)
  (set! *LINK_UNIQUE*   1)
  (set! *TRIANG_UNIQUE* 1)
  (set! *QUAD_UNIQUE*   1)
  (set! *BOX_UNIQUE*    1)
  (set! *GROUP_UNIQUE*  1)
  (set! *POINTS*  '())
  (set! *LINKS*   '())
  (set! *TRIANGS* '())
  (set! *QUADS*   '())
  (set! *BOXES*   '())
  (set! *GROUPS*  '())
  (set! *POINT_MAX* 1.0)
  ; remove undo operations as well
  (undo-redo-reset)
)


(define (object-valid obj)
  (case (car obj)
    ( (p) (point-valid-basic  (cadr obj)) )
    ( (l) (link-valid-basic   (cadr obj)) )
    ( (t) (triang-valid-basic (cadr obj)) )
    ( (q) (quad-valid-basic   (cadr obj)) )
    ( (b) (box-valid-basic    (cadr obj)) )
    ( (g) (group-valid-basic  (cadr obj)) )
    (else #f)
  )
)


; "n" is an object descriptor e.g.: '(p 4) or '(l 45)
(define (object-property-print n)
  (case (car n)
    ( (p)
      (let* ((coord (object-property-get 'p (cadr n) 'coords)))
        (string-append "Point " (number->string (cadr n))
                       " --- X: " (number->string (car coord))
                       " Y: "     (number->string (cadr coord))
                       " Z: "     (number->string (caddr coord))
        )
      )
    )
    ( (l)
      (let* ((nodes (object-property-get 'l (cadr n) 'nodes)))
        (string-append "Link " (number->string (cadr n))
                       " --- Nodes: " 
                       (number->string (car nodes)) " "
                       (number->string (cadr nodes))
        )
      )
    )
    ( (t)
      (let* ((nodes (object-property-get 't (cadr n) 'nodes)))
        (string-append "Triang " (number->string (cadr n))
                       " --- Nodes: " 
                       (number->string (car nodes)) " "
                       (number->string (cadr nodes)) " "
                       (number->string (caddr nodes))
        )
      )
    )
    ( (q)
      (let* ((nodes (object-property-get 'q (cadr n) 'nodes)))
        (string-append "Quad " (number->string (cadr n))
                       " --- Nodes: " 
                       (number->string (car nodes)) " "
                       (number->string (cadr nodes)) " "
                       (number->string (caddr nodes)) " "
                       (number->string (cadddr nodes))
        )
      )
    )
    ( (g)
      (let* ()
        (string-append "Group " (number->string (cadr n)))
      )
    )
    (else "")
  )
)


; --------------------------------------------------------
; Points
; --------------------------------------------------------

; This function validates a point identification number,
; - it should be a number,
; - an element with that ID should exist and
; - it should not be a part of a group
(define (point-valid-basic id)
  (and (number? id)
       (assoc id *POINTS*)
       (not (object-property-get 'p id 'group))
  )
)

; This function adds a new point to the list of points.
; id - point identifier, it can be #f then it is taken from *POINT_UNIQUE*
; coords - a list of three numbers, coordinates of the point
; properties - assoc list of other properties or #f
; undo - #t or #f, whether undo is necessary
; It returns the identification number of the created point.
(define (point-add-basic id coords properties undo)
  ; determine the ID of a new point
  (if (not id)
    (begin
      (set! id *POINT_UNIQUE*)
      ; store the last created object
      (set! *OBJECT_LAST* (list 'p *POINT_UNIQUE*))
      ; increase the unique identifier
      (set! *POINT_UNIQUE* (1+ *POINT_UNIQUE*))
    )
  )
  ; add the point to the list
  (set! *POINTS* (cons (list id) *POINTS*))
  ; add coordinates to the point
  (object-property-set 'p id 'coords coords #t)
  ; node-element connectivity is set to default, but maybe overwritten
  (object-property-set 'p id 'parent '() #t)
  ; if some other properties are also specified set them as well 
  ; except coordinates
  (if properties
    (begin
      (set! properties (assoc-remove! properties 'coords))
      (for-each 
        (lambda (x) (object-property-set 'p id (car x) (cdr x) #t))
        properties
      )
    )
  )

  ; save the undo or redo information
  (if undo
    (undo-add (list point-del-basic id #f))
    (redo-add (list point-del-basic id #t))
  )

  ; maximum extend of the points
  (if (< *POINT_MAX* (abs (car coords)))
    (set! *POINT_MAX* (abs (car coords)))
  )
  (if (< *POINT_MAX* (abs (cadr coords)))
    (set! *POINT_MAX* (abs (cadr coords)))
  )
  (if (< *POINT_MAX* (abs (caddr coords)))
    (set! *POINT_MAX* (abs (caddr coords)))
  )

  ; increase the selection buffer
  (set! *GL_SELECT_BUFFER*
    (make-uniform-array 100 (* 5 (+ (length *POINTS*)
                                    (length *LINKS*)
                                    (length *TRIANGS*)
                                    (length *QUADS*)
                                    (length *BOXES*)
                                    (length *GROUPS*)))
    )
  )

  id
)

; This function deletes a point from the list of points.
; n - position of point in the list to delete
; undo - #t or #f, whether undo is necessary
(define (point-del-basic n undo)
  ; check that there is/are point(s) to delete
  (if (not (null? *POINTS*))
    ; determine that the point exist
    (if (assoc n *POINTS*)
      (begin
        ; no other element should be connected
        (if (null? (object-property-get 'p n 'parent))
          (begin
            ; here we are abusing the system a little bit,
            ; the explicitly specified coordinates are 0 0 0
            ; but the real coordinates are in the property list !!!
            (if undo
              (undo-add (list point-add-basic n 
                              (list 'quote (object-property-get 'p n 'coords))
                              (list 'quote (assoc-ref *POINTS* n))
                              #f))
              (redo-add (list point-add-basic n
                              (list 'quote (object-property-get 'p n 'coords))
                              (list 'quote (assoc-ref *POINTS* n))
                              #t))
            )
            ; remove the point with its properties
            (set! *POINTS* (assoc-remove! *POINTS* n))
          )
          (begin
            (gui-entry-add-history (string-append
                                      "\nObjects are still conntected to point "
                                      (number->string n)
                                      " -> point is not deleted"
                                   ))
          )
        )
      )
      (begin
        (gui-entry-add-history (string-append "Point " 
                                              (number->string n)
                                              " does not exist"))
      )
    )
  )
)

; ---------------------------------------------------------------------
; Element handling macros
; These macros are applicable to link, triangle, quad and box elements
; ---------------------------------------------------------------------

; validation functions definition
; -------------------------------
(define-macro (create-validate func-name type-letter elem-list)

  `(define (,func-name id)
    (and (number? id)
         (assoc id ,elem-list)
         (not (object-property-get ',type-letter id 'group))
    )
  )

)

; adding function definition
; --------------------------
(define-macro (create-add func-name func-del
                          type-letter elem-unique elem-list)

  `(define (,func-name id nodes properties undo)
    ; create new ID
    (if (not id)
      (begin
        (set! id ,elem-unique)
        ; store the last created object
        (set! *OBJECT_LAST* (list ',type-letter ,elem-unique))
        ; increase the unique identifier
        (set! ,elem-unique (1+ ,elem-unique))
      )
    )

    ; add the element to the list
    (set! ,elem-list (cons (list id) ,elem-list))
    ; add nodes property
    (object-property-set ',type-letter id 'nodes nodes #t)
    ; set default child, parent relation
    (object-property-set ',type-letter id 'child  '() #t)
    (object-property-set ',type-letter id 'parent '() #t)
    ; if some other properties are also specified set them, except nodes
    (if properties
      (begin
        (set! properties (assoc-remove! properties 'nodes))
        (for-each 
          (lambda (x) (object-property-set ',type-letter id (car x) (cdr x) #t))
          properties
        )
      )
    )

    ; set the reference with the points
    (for-each 
      (lambda (x)
        (object-property-set 'p x 'parent
                             (cons (list ',type-letter id) 
                                   (object-property-get 'p x 'parent)
                             )
                             #t)
      )
      nodes
    )
    ; set the reference with any other children
    (for-each 
      (lambda (x)
        (object-property-set (car x) (cadr x) 'parent
                             (cons (list ',type-letter id) 
                                   (object-property-get (car x) (cadr x) 
                                                        'parent)
                             )
                             #t)
      )
      (object-property-get ',type-letter id 'child)
    )
    
    ; set undo operation
    (if undo
      (undo-add (list ,func-del id #f))
      (redo-add (list ,func-del id #t))
    )

    ; increase the selection buffer
    (set! *GL_SELECT_BUFFER*
      (make-uniform-array 100 (* 5 (+ (length *POINTS*)
                                      (length *LINKS*)
                                      (length *TRIANGS*)
                                      (length *QUADS*)
                                      (length *BOXES*)
                                      (length *GROUPS*)))
      )
    )
    id
  )

)


; deleting function definition
; ----------------------------
(define-macro (create-del func-name func-add
                          type-letter elem-list)

  `(define (,func-name id undo)
    ; anything to delete
    (if (not (null? ,elem-list))
      ; does it exist ?
      (if (assoc id ,elem-list)
        ; no parent element is connected ?
        (if (null? (object-property-get ',type-letter id 'parent))
          (let* 
            ( (nodes (object-property-get ',type-letter id 'nodes))
              (child (object-property-get ',type-letter id 'child))
            )
            (if undo
              (undo-add (list ,func-add id 
                              (list 'quote nodes) 
                              (list 'quote (assoc-ref ,elem-list id))
                              #f))
              (redo-add (list ,func-add id
                              (list 'quote nodes)
                              (list 'quote (assoc-ref ,elem-list id))
                              #t))
            )
            ; remove the element from the list
            (set! ,elem-list (assoc-remove! ,elem-list id))
            ; remove references to this element from the points
            (for-each
              (lambda (x)
                (object-property-set 'p x 'parent
                  (delete (list ',type-letter id)
                          (object-property-get 'p x 'parent)) #t)
              )
              nodes
            )
            ; remove references from any child
            (for-each 
              (lambda (x)
                (object-property-set (car x) (cadr x) 'parent
                  (delete (list ',type-letter id)
                          (object-property-get (car x) (cadr x) 'parent)) #t)
              )
              child
            )
          )
          (begin
            (gui-entry-add-history (string-append
                                      "\nObjects are still conntected to "
                                      (symbol->string ',type-letter) " "
                                      (number->string id)
                                      " -> object is not deleted"))
          )
        )
        (begin
          (gui-entry-add-history (string-append
                                    "\nObject "
                                    (symbol->string ',type-letter) " "
                                    (number->string id)
                                    " does not exist") )
        )
      )
    )
  )

)


; --------------------------------------------------------
; Links
; --------------------------------------------------------
(create-validate link-valid-basic l *LINKS*)
(create-add      link-add-basic link-del-basic l *LINK_UNIQUE* *LINKS*)
(create-del      link-del-basic link-add-basic l *LINKS*)

; --------------------------------------------------------
; Triangles
; --------------------------------------------------------
(create-validate triang-valid-basic t *TRIANGS*)
(create-add      triang-add-basic triang-del-basic t *TRIANG_UNIQUE* *TRIANGS*)
(create-del      triang-del-basic triang-add-basic t *TRIANGS*)

; --------------------------------------------------------
; Quadrilaterals
; --------------------------------------------------------
(create-validate quad-valid-basic q *QUADS*)
(create-add      quad-add-basic quad-del-basic q *QUAD_UNIQUE* *QUADS*)
(create-del      quad-del-basic quad-add-basic q *QUADS*)

; --------------------------------------------------------
; Boxes
; --------------------------------------------------------
(create-validate box-valid-basic b *BOXES*)
(create-add      box-add-basic box-del-basic b *BOX_UNIQUE* *BOXES*)
(create-del      box-del-basic box-add-basic b *BOXES*)

; --------------------------------------------------------
; Groups
; --------------------------------------------------------

; validate the group number, 
(define (group-valid-basic id)
  (and (number? id)
       (assoc id *GROUPS*)
       (not (object-property-get 'g id 'group))
  )
)


(define (group-create-basic id selection properties undo)
  ; determine the ID of a new group
  (if (not id)
    (begin
      (set! id *GROUP_UNIQUE*)
      ; increase the unique identifier
      (set! *GROUP_UNIQUE* (1+ *GROUP_UNIQUE*))
    )
  )
  ; add the group to the list
  (set! *GROUPS* (cons (list id) *GROUPS*))

  (object-property-set 'g id 'parts selection #t)
  ; if some other properties are also specified set them as well except
  ; parts
  (if properties
    (begin
      (set! properties (assoc-remove! properties 'parts))
      (for-each
        (lambda (x) (object-property-set 'g id (car x) (cdr x) #t))
        properties
      )
    )
  )

  ; set that the points/elements belong to the new group
  ; points can belong to several groups,
  ; elements belong only to one group.
  (for-each 
    (lambda (x)
      (case (car x)
        ( (p)
          (let* ( (group (object-property-get 'p (cadr x) 'group)) )
            (if group
              (object-property-set 'p (cadr x) 'group (cons id group) #t)
              (object-property-set 'p (cadr x) 'group (list id) #t)
            )
          )
        )
        ( else
          (object-property-set (car x) (cadr x) 'group id #t)
        )
      )
    )
    selection
  )

  ; save the undo or redo information
  (if undo
    (undo-add (list group-explode-basic id #f))
    (redo-add (list group-explode-basic id #t))
  )

  ; increase the selection buffer
  (set! *GL_SELECT_BUFFER*
    (make-uniform-array 100 (* 5 (+ (length *POINTS*)
                                    (length *LINKS*)
                                    (length *TRIANGS*)
                                    (length *QUADS*)
                                    (length *BOXES*)
                                    (length *GROUPS*)))
    )
  )

  id
)


(define (group-explode-basic id undo)
  ; anything to delete
  (if (not (null? *GROUPS*))
    ; does it exist ?
    (if (assoc id *GROUPS*)
      (let* ( (parts (object-property-get 'g id 'parts)) )
        (if parts
          (begin
            (if undo
              (undo-add (list group-create-basic id
                              (list 'quote parts)
                              (list 'quote (assoc-ref *GROUPS* id))
                              #f))
              (redo-add (list group-create-basic id
                              (list 'quote parts)
                              (list 'quote (assoc-ref *GROUPS* id))
                              #f))
            )
            ; remove the group
            (set! *GROUPS* (assoc-remove! *GROUPS* id))
            ; remove references to the group
            (for-each 
              (lambda (x)
                (case (car x)
                  ( (p)
                    (let* ( (group (object-property-get 'p (cadr x) 'group)) )
                      (if (= (length group) 1)
                        (object-property-del 'p (cadr x) 'group #t)
                        (object-property-set 'p (cadr x) 'group 
                                             (delete id group) #t)
                      )
                    )
                  )
                  ( else
                    (object-property-del (car x) (cadr x) 'group #t)
                  )
                )
              )
              parts
            )
          )
        )
      )
      (begin
        (gui-entry-add-history (string-append 
                                  "\nGroup " 
                                  (number->string id)
                                  " does not exist"))
      )
    )
  )
)


; returns all points which are referenced inside the group
(define (group-get-points-basic id)
  (let* ( (plist '()) )
    (for-each (lambda (x)
      (case (car x)
        ( (p) 
          (set! plist (cons x plist)) )
        ( (g)
          (set! plist (append (group-get-points-basic (cadr x)) plist)) )
      ) )
      (object-property-get 'g id 'parts)
    )
    plist
  )
)

; returns all elements which are referenced inside the group
(define (group-get-objects-basic id)
  (let* ( (plist '()) )
    (for-each (lambda (x)
      (case (car x)
        ( (g)
          (set! plist (append (group-get-objects-basic (cadr x)) plist)) )
        ( else (set! plist (cons x plist)) )
      ) )
      (object-property-get 'g id 'parts)
    )
    plist
  )
)

;---------------------------------------------------------


