(define-module (sx gui-entry))
(use-modules (gtk gdk) (gtk gtk))
(export *ENTRY_LABEL* *ENTRY_TEXT* *ENTRY_HISTORY* *ENTRY_DEFAULT_TEXT*
        *ENTRY_CALLBACK* *ENTRY_SIZE*
        *COMMAND_CACHE*
        text-split-with-empty
        entry-add-command-list
        entry-add-abbrev
        gui-entry-callback-execute 
        gui-entry-callback-replace
        gui-entry-default-callback
;;;
        gui-entry-label-set-text gui-entry-label-get-text 
        gui-entry-set-text gui-entry-get-text
        gui-entry-add-history gui-entry-execute
)


; widget for text in front of the entry
(define *ENTRY_LABEL*     #f)
; the entry widget, typed commands come here
(define *ENTRY_TEXT*      #f)
; all commands come to this widget, not editable
(define *ENTRY_HISTORY*   #f)

(define *ENTRY_DEFAULT_TEXT* "Command: ")

; storage of the callback function to execute when
; the entry is activated (ENTER or space is pressed)
(define *ENTRY_CALLBACK*  #f)

; the size of the entry is limited,
; to avoid overloading it when cross selection is used
; the selection must be specified in chunks
; this number determines how many list elements is converted
; and added to the selection at once
(define *ENTRY_SIZE*  50)

; abbreviated commands
(define *COMMAND-ABBREV* 
  '( ("p"  "point")
     ("l"  "link")
     ("d"  "delete")
     ("m"  "move")
     ("c"  "copy")
     ("s"  "select")
     ("q"  "quit")
     ("r"  "redo")
     ("e"  "edit")
     ("z"  "zoom")
   )
)

; the list of available commands
(define *COMMANDS-LIST*
  '( "point"  "link" "triang-point" "triang-side" "quad-point" "quad-side" "box"
     "group" "ungroup"
     "delete" "move" "copy" "select" "divide"
     "edit" "display" "getcoord" "getdist"
     "reset" "center" "style" "light" "zoom"
     "u" "undo" "redo"
     "new" "save" "open" "quit"
   )
)


; This variable can store a series of commands which are executed
; after another. Using this variable the menus can issues several
; consequtive commands
(define *COMMAND_CACHE* '())


;(define *COMMAND_CACHE* '("p" "0" "0.2" "0.2,0.2" "0,0.2" "" 
;  "c" "all" "" "@0.1,0.1,-0.1" "triang" "1" "2" "3" "" 
;  "quad" "5" "6" "7" "8" ""))


; private variable for the last executed command
(define *COMMAND_LAST* #f)


;;; --------------------------------------------------------------
;;; gui independent part
;;; --------------------------------------------------------------

; This is based on string-split in guile 1.5, but it
; removes empty strings from the list when "empty" is #f
(define (text-split-with-empty str ch empty)
  (let* 
    ((idx (string-length str))
     (last #f)
     (slist '())
    )
    (while (>= idx 0)
      (set! last idx)
      (while (and (> idx 0) (not (char=? (string-ref str (1- idx)) ch)))
        (set! idx (1- idx))
      )
      (if (>= idx 0)
        (begin
          (if (or empty
                  (and (not empty) (> (- last idx) 0)) )
            (set! slist (cons (substring str idx last) slist))
          )
          (set! idx (1- idx))
        )
      )
    )
    slist
  )
)

;;; --------------------------------------------------------------

; adds a list of commands to the list of available user commands
(define (entry-add-command-list command)
  (set! *COMMANDS-LIST* (append *COMMANDS-LIST* command))
)

; adds an abbreviation
(define (entry-add-abbrev abbrev)
  (set! *COMMAND-ABBREV* (cons abbrev *COMMAND-ABBREV*))
)


; the callback function which executed by the entry widget
(define (gui-entry-callback-execute)
  (*ENTRY_CALLBACK*)
  (if (not (null? *COMMAND_CACHE*))
    (begin
      (for-each (lambda (x)
          (gui-entry-set-text x)
          (*ENTRY_CALLBACK*)
        )
        *COMMAND_CACHE*
      )
      (set! *COMMAND_CACHE* '())
    )
  )
  ; finally clear the entry
  (gui-entry-set-text "")
)


; replaces the default callback
(define (gui-entry-callback-replace callback-function)
  (set! *ENTRY_CALLBACK* callback-function)
)


; the default entry callback function
(define (gui-entry-default-callback)
  (let* 
    ( (pp (gui-entry-get-text)) )
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text)))

    (if (= (string-length pp) 0)
      (if *COMMAND_LAST*
        (begin
          (gui-entry-add-history *COMMAND_LAST*)
          ; execute last command
          ; extra parenthesis is necessary to execute it !!!
          ((primitive-eval 
               (string->symbol (string-append "command-" *COMMAND_LAST*))))
        )
      )
      (let*
        ( (func '()) 
          (abbrev (assoc pp *COMMAND-ABBREV*))
          (mesg "")
        )

        (if abbrev
          ; if there is an abbreviation for this command use that
          (begin
            (set! func (cdr abbrev))
          )
          ; try exact match
          (if (member pp *COMMANDS-LIST*)
            (set! func (list pp))
            ; otherwise try to match a partial command
            (for-each 
              (lambda (com)
                (if (and (<= (string-length pp) (string-length com))
                         (string-ci=? pp (substring com 0 (string-length pp))) )
                  (set! func (cons com func))
                )
              )
              *COMMANDS-LIST*
            )
          )
        )

        (case (length func)
          ( (0)
            (gui-entry-add-history (string-append pp " *Unknown command*"))
          )
          ( (1)
            (gui-entry-add-history (car func))
            (set! *COMMAND_LAST* (car func)) ; store it as last command
            ; execute the command
            ; extra parenthesis is necessary to execute it !!!
            ((primitive-eval 
                 (string->symbol (string-append "command-" (car func)))))
          )
          ( else
            ; several commands matches
            (set! mesg (string-append "\n*Ambiguous command was " pp " for [ "))
            (for-each 
              (lambda (xx)
                (set! mesg (string-append mesg xx " "))
              )
              func
            )
            (set! mesg (string-append mesg "]*"))
            (gui-entry-add-history mesg)
          )
        )
      )
    )
  )
)


;;; --------------------------------------------------------------
;;; GTK+ based functions
;;; --------------------------------------------------------------

; set the text of the entry for instructions to the user
(define (gui-entry-label-set-text text)
  (gtk-label-set-text *ENTRY_LABEL* text)
)

(define (gui-entry-label-get-text)
  (gtk-label-get *ENTRY_LABEL*)
)

(define (gui-entry-set-text text)
  (gtk-entry-set-text *ENTRY_TEXT* text)
)

; read back the users answer
(define (gui-entry-get-text)
  (let*
    ( (tt (gtk-entry-get-text *ENTRY_TEXT*)) ; store the entry
      ; position of last character, it can be an empty string
      (ll (if (> (string-length tt) 0) (1- (string-length tt)) 0))
    )
    (gtk-entry-set-text *ENTRY_TEXT* "")     ; clear the entry
    ; if there is a space at the end of the string then remove it
    (if (string=? (substring tt ll) " ")
      (substring tt 0 ll)
      tt                                       ; return the value
    )
  )
)


(define (gui-entry-add-history text)
  (gtk-text-insert *ENTRY_HISTORY* #f "black" "white" text -1)
)


(define (gui-entry-execute command)
  ; keep focus at the entry
  (gtk-widget-grab-focus *ENTRY_TEXT*)
  ; execute the required command
  (gtk-entry-set-text *ENTRY_TEXT* command)
  (gui-entry-callback-execute)
;  (gtk-signal-emit *ENTRY_TEXT* "activate")
)


