(define-module (sx mouse))
(use-modules (sx render) (sx trackball) (sx selection-basic)
             (sx gui-gl) (sx gui-entry) (sx gui-selection))
(export list->selection
        GL-button1-dragging GL-button2-dragging GL-button3-dragging
        GL-button-press-callback 
        GL-button-motion-callback
        GL-button-release-callback)

;;; Graphics/OpenGL window handling
;;; ---------------------------------------------------------------------

; the starting position where the mouse button has been pressed
(define *STARTX* 0)
(define *STARTY* 0)

; size of the picking cursor
(define *PICK_SIZE* 7)

; the handler function when the mouse is moving with a button pressed
(define *GL_DRAGGING_HANDLER* #f)

; converts the selected list of objects into a textual command
; which can be passed to the command processor
(define (list->selection sellist)
  (let*
    ((selection ""))
    (for-each
      (lambda (x)
        (set! selection 
          (string-append selection
                         (case (car x)
                           ((p) ",point,")
                           ((l) ",link,")
                           ((t) ",triangle,")
                           ((q) ",quad,")
                           ((b) ",box,")
                           ((g) ",group,") )
                         (number->string (cadr x))))
      )
      sellist
    )
    selection
  )
)


; panning 
(define (GL-button1-dragging x y w h)
  (set! *XDIST* (+ *XDIST* (* (- x *STARTX*) (/ 0.05 *ZOOM*)) ))
  (set! *YDIST* (- *YDIST* (* (- y *STARTY*) (/ 0.05 *ZOOM*)) ))

  (set! *STARTX* x)
  (set! *STARTY* y)
)

; rotating
(define (GL-button2-dragging x y w h)
  (let*
    (
     (q (trackball-calc-quaternion 
          (/ (- (* 2.0 *STARTX*) w) w)
          (/ (- h (* 2.0 *STARTY*)) h)
          (/ (- (* 2.0 x) w) w)
          (/ (- h (* 2.0 y)) h)))
    )
    (set! *QUATERNION* (trackball-add-quaternions q *QUATERNION*))
  )
  (set! *STARTX* x)
  (set! *STARTY* y)
)

; zooming
(define (GL-button3-dragging x y w h)
  (let*
    ((dy (- y *STARTY*))
     (sn (if (> dy 0) 1.05 0.95))
    )
    (set! *ZOOM* (* sn *ZOOM*))
  )
  (set! *STARTX* x)
  (set! *STARTY* y)
)


(define (GL-button-press-callback x y pressed_button)
  (case *PICK_MODE*
    ; only viewing
    ( (view)
      (set! *STARTX* x)
      (set! *STARTY* y)
      (set! *GL_DRAGGING_HANDLER* pressed_button)
    )
    ; selecting _one_ point only, and using actually the coordinates of it
    ( (select_coord)
      (case pressed_button
        ( (1)
          (let*
            ; select _one_ point from the screen
            ( (coord (gui-selection-single-coord
                         (render-scene 'select x y *PICK_SIZE* *PICK_SIZE* #f))) )
            (if coord
              ; FIXME: we should actually check, that nothing has been typed in
              ; only filters like: .x .xy
              ; if there is more, no selection at all
              (gui-entry-execute
                (string-append (gui-entry-get-text) ","
                               (number->string (car coord)) ","
                               (number->string (cadr coord)) ","
                               (number->string (caddr coord)))
              ) ; add the selection with prefixes
            )
          )
        )
        ( (3) ; stop selection
          (gui-entry-execute "")
        )
      )
    )
    ; selecting one object by picking
    ( (select_obj)
      (case pressed_button
        ( (1) ; try to select the object below or switch to crossing
          (let*
            ((sel (render-scene 'select x y *PICK_SIZE* *PICK_SIZE* #f)))
            (if sel
              ; add the selection through command-line
              ; here we may overload the buffer of the entry,
              ; so we have to supply the selection in chunks
              (let*
                ( (new-sel sel) )
                (while (> (length new-sel) *ENTRY_SIZE*)
                  (gui-entry-execute 
                    (list->selection (list-head new-sel *ENTRY_SIZE*)))
                  (set! new-sel (list-tail new-sel *ENTRY_SIZE*))
                )
                (gui-entry-execute (list->selection new-sel))
              )
              (begin ; switch to crossing selection mode
                (selection-crossing-enter "Specify opposite corner: " 
                                          (list x y))
              )
            )
          )
        )
        ( (3) ; stop selection
          (gui-entry-execute "")
        )
      )
    )
    ; selecting by cross
    ( (select_cross)
      (case pressed_button
        ( (1) ; first button activates it
          (if *SELECT_WIN_CORNER_1*
            (let*
              ((cx (/ (+ (car *SELECT_WIN_CORNER_1*)  x) 2.0))
               (cy (/ (+ (cadr *SELECT_WIN_CORNER_1*) y) 2.0))
               (sx (abs (- x (car *SELECT_WIN_CORNER_1*))))
               (sy (abs (- y (cadr *SELECT_WIN_CORNER_1*))))
               (sel #f)
              )
              ; reset to single pick
              (selection-crossing-quit)
              ; remove the selecting box
              ; if some objects are selected, then the screen is updated twice,
              ; here to remove the box quickly and 
              ; inside selection to show the newly selected objects
              (gui-update-gl-window #f)

              ; check for selection only if the cursor is moved
              (if (or (> sx 0) (> sy 0))
                (begin
                  (set! sel (render-scene 'select cx cy sx sy #f))
                  (if sel
                    ; add the selection,
                    ; here we may overload the buffer of the entry,
                    ; so we have to supply the selection in chunks
                    (let*
                      ( (new-sel sel) )
                      (while (> (length new-sel) *ENTRY_SIZE*)
                        (gui-entry-execute 
                          (list->selection (list-head new-sel *ENTRY_SIZE*)))
                        (set! new-sel (list-tail new-sel *ENTRY_SIZE*))
                      )
                      (gui-entry-execute (list->selection new-sel))
                    )
                  )
                )
              )
            )
            (begin
              ; no point has been selected, so this is the first one
              (gui-entry-label-set-text "Specify opposite corner: ")
              (set! *SELECT_WIN_CORNER_1* (list x y))
            )
          )
        )
        ( (3) ; third button cancels crossing selection
          (selection-crossing-quit)
          ; remove the selecting box
          (gui-update-gl-window #f)
        )
      )
    )
  )
)


(define (GL-button-motion-callback x y)
  (case *PICK_MODE*
    ( (view)
      (case *GL_DRAGGING_HANDLER*
        ( (1) 
          (GL-button1-dragging x y *GL_VIEWPORT_WIDTH* *GL_VIEWPORT_HEIGHT*)
          (gui-update-gl-window #f)
        )
        ( (2) 
          (GL-button2-dragging x y *GL_VIEWPORT_WIDTH* *GL_VIEWPORT_HEIGHT*)
          (gui-update-gl-window #f)
        )
        ( (3) 
          (GL-button3-dragging x y *GL_VIEWPORT_WIDTH* *GL_VIEWPORT_HEIGHT*)
          (gui-update-gl-window #f)
        )
      )
    )
    ( (select_cross)
      (if *SELECT_WIN_CORNER_1*
        (begin
          (set! *SELECT_WIN_CORNER_2* (list x y))
          (gui-update-gl-window #f)
        )
      )
    )
  )
)


(define (GL-button-release-callback released_button x y)
  (if (eq? released_button *GL_DRAGGING_HANDLER*)
    (set! *GL_DRAGGING_HANDLER* #f)
  )
)





