(define-module (sx property))
(use-modules (sx undo) (sx gui-entry))
(export object-property-set object-property-del object-property-get
        object-property-get-all-user
        object-property-collect
        string->object-property object-property->string
        object-property-undo-del object-property-undo-set
        object-property-undo-update)

; maximum extend of the points
(define-public *POINT_MAX* 1.0)

; the list of points, it is an association list
; an element of the list is (n (...))
; where n is the identification number of the point
; and (...) is an association list of properties
; 0 - coordinates in the form of (0 (x y z))
(define-public *POINTS* '())
(define-public *LINKS*   '())
(define-public *TRIANGS* '())
(define-public *QUADS*   '())
(define-public *BOXES*   '())
(define-public *GROUPS*  '())


; Each point has a unique identification number.
; This number can only be increased !!!
(define-public *POINT_UNIQUE* 1)
(define-public *LINK_UNIQUE* 1)
(define-public *TRIANG_UNIQUE* 1)
(define-public *QUAD_UNIQUE* 1)
(define-public *BOX_UNIQUE* 1)
(define-public *GROUP_UNIQUE* 1)

; the last created object is stored here
(define-public *OBJECT_LAST* #f)

; OpenGL buffer for element selection
(define-public *GL_SELECT_BUFFER* (make-uniform-array 100 5))


;;; property handling functions
;;; ------------------------------------------------------------------

; This variable stores the internal property names. These properties
; should not be modified by the user explicitly, only through provided
; functions.
(define *PROP_INTERNAL* '(coords nodes group parts child parent))


; returns #t when it could carry out the operation and
; #f when there was an error
(define (object-property-set type id prop_id prop allow-internal)
  (if (or allow-internal 
          (not (or allow-internal (member prop_id *PROP_INTERNAL*))))
    (let* 
      ( (elem #f) )
      (case type
        ( (p) (set! elem (assoc id *POINTS*)) )
        ( (l) (set! elem (assoc id *LINKS*)) )
        ( (t) (set! elem (assoc id *TRIANGS*)) )
        ( (q) (set! elem (assoc id *QUADS*)) )
        ( (b) (set! elem (assoc id *BOXES*)) )
        ( (g) (set! elem (assoc id *GROUPS*)) )
      )
      (if elem
        (begin
          (set! elem (assoc-set! (cdr elem) prop_id prop))
          (case type
            ( (p) (set! *POINTS*  (assoc-set! *POINTS*  id elem)) #t)
            ( (l) (set! *LINKS*   (assoc-set! *LINKS*   id elem)) #t)
            ( (t) (set! *TRIANGS* (assoc-set! *TRIANGS* id elem)) #t)
            ( (q) (set! *QUADS*   (assoc-set! *QUADS*   id elem)) #t)
            ( (b) (set! *BOXES*   (assoc-set! *BOXES*   id elem)) #t)
            ( (g) (set! *GROUPS*  (assoc-set! *GROUPS*  id elem)) #t)
          )
          #t
        )
        (begin
          (gui-entry-add-history 
            (string-append "\n*Object " 
                           (symbol->string type) " "
                           (number->string id)
                           " does not exist*"))
          #f
        )
      )
    )
  )
)


; returns #t when it could carry out the operation and
; #f when there was an error
(define (object-property-del type id prop_id allow-internal)
  (if (or allow-internal 
          (not (or allow-internal (member prop_id *PROP_INTERNAL*))))
    (let* 
      ( (elem #f) )
      (case type
        ( (p) (set! elem (assoc id *POINTS*)) )
        ( (l) (set! elem (assoc id *LINKS*))  )
        ( (t) (set! elem (assoc id *TRIANGS*))  )
        ( (q) (set! elem (assoc id *QUADS*))  )
        ( (b) (set! elem (assoc id *BOXES*))  )
        ( (g) (set! elem (assoc id *GROUPS*)) )
      )
      (if elem
        (begin
          (set! elem (assoc-remove! (cdr elem) prop_id))
          (case type
            ( (p) (set! *POINTS*  (assoc-set! *POINTS*  id elem)) #t)
            ( (l) (set! *LINKS*   (assoc-set! *LINKS*   id elem)) #t)
            ( (t) (set! *TRIANGS* (assoc-set! *TRIANGS* id elem)) #t)
            ( (q) (set! *QUADS*   (assoc-set! *QUADS*   id elem)) #t)
            ( (b) (set! *BOXES*   (assoc-set! *BOXES*   id elem)) #t)
            ( (g) (set! *GROUPS*  (assoc-set! *GROUPS*  id elem)) #t)
          )
          #t
        )
        (begin
          (gui-entry-add-history 
            (string-append "\n*Object " 
                           (symbol->string type) " "
                           (number->string id)
                           " does not exist*"))
          #f
        )
      )
    )
  )
)

;; returns the property with the given ID
(define (object-property-get type id prop_id)
  (let* 
    ( (elem #f) (prop #f) )
    (case type
      ( (p) (set! elem (assoc id *POINTS*)) )
      ( (l) (set! elem (assoc id *LINKS*)) )
      ( (t) (set! elem (assoc id *TRIANGS*)) )
      ( (q) (set! elem (assoc id *QUADS*)) )
      ( (b) (set! elem (assoc id *BOXES*)) )
      ( (g) (set! elem (assoc id *GROUPS*)) )
    )
    (if elem
      (begin
        (assoc-ref (cdr elem) prop_id)
      )
      (begin
        (gui-entry-add-history 
          (string-append "\n*Object " 
                         (symbol->string type) " "
                         (number->string id)
                         " does not exist*"))
        #f
      )
    )
  )
)

;; returns all non-internal, user defined properties
(define (object-property-get-all-user type id)
  (let* 
    ( (elem #f) (prop-list '()) )
    (case type
      ( (p) (set! elem (assoc id *POINTS*)) )
      ( (l) (set! elem (assoc id *LINKS*)) )
      ( (t) (set! elem (assoc id *TRIANGS*)) )
      ( (q) (set! elem (assoc id *QUADS*)) )
      ( (b) (set! elem (assoc id *BOXES*)) )
      ( (g) (set! elem (assoc id *GROUPS*)) )
    )
    (if elem
      (begin
        (for-each
          (lambda (prop)
            (if (not (member (car prop) *PROP_INTERNAL*))
              (set! prop-list (cons prop prop-list))
            )
          )
          (cdr elem)
        )
        prop-list
      )
      (begin
        (gui-entry-add-history 
          (string-append "\n*Object " 
                         (symbol->string type) " "
                         (number->string id)
                         " does not exist*"))
        #f
      )
    )
  )
)


;; from a list of element definitions; e.g.: '((l 1) (p 4) (q 5))
;; it collects all non-internal properties into an association list;
;; the elements of the created association list are converted to 
;; strings so they can be displayed in a widget of the user interface
(define (object-property-collect elem-list)
  (let* ( (coll-list '()) 
          (property #f) (type #f)
        )
    ; go through all objects
    (for-each
      (lambda (obj)
        ; go through all properties in the object
        (for-each
          (lambda (y)
            ; check whether we have already stored this property
            (set! property (assoc (car y) coll-list))
            (if property
              (begin
                ; check whether we have this type of property
                (set! type (assoc (cadr y) (cdr property)))
                (if type
                  ; type exist, if this value is not in the list, store it
                  (if (not (member (caddr y) (cdr type)))
                    (begin
                      (set! property 
                            (assoc-set! (cdr property) 
                                        (cadr y) 
                                        (cons (caddr y) (cdr type))))
                      ; store the new property collection back
                      (set! coll-list (assoc-set! coll-list (car y) property))
                    )
                  )
                  ; this type does not exist
                  (begin
                    (set! property (cons (cdr y)
                                         (cdr property)))
                    ; store the new property collection back
                    (set! coll-list (assoc-set! coll-list (car y) property))
                  )
                )
              )
              ; new property
              (begin
                (set! coll-list (cons (list (car y) (cdr y))
                                      coll-list))
              )
            )
          )
          ; user properties
          (object-property-get-all-user (car obj) (cadr obj))
        )
      )
      elem-list
    )

    ; now convert everything to strings
    (map
      (lambda (prop)
        (cons (symbol->string (car prop))
              (map
                (lambda (x)
                  (cons (symbol->string (car x))
                        (map 
                          (lambda (y) (object-property->string (car x) y))
                          (cdr x)
                        )
                  )
                )
                (cdr prop)
              )
        )
      )
      coll-list
    )
  )
)


;; converts a string (which is the user entered property value)
;; to a scheme object, strings, integers, reals or lists
;; if the data is invalid it returns #f
;; type is a symbol !!!
(define (string->object-property type string)
  (case type
    ( (word) (if (> (string-length string) 0) string #f))
    ( (integer)
      (if (integer? (string->number string))
        (string->number string)
        #f
      )
    )
    ( (real) (string->number string))
    ( (word-list) 
      (let* ((ll (text-split-with-empty string #\, #f)))
        (if (null? ll) #f ll)
      )
    )
    ( (integer-list)
      (call-with-current-continuation
        (lambda (return)
          (map
            (lambda (x) 
              (if (not (integer? (string->number x)))
                (return #f)
              )
              (string->number x)
            )
            (text-split-with-empty string #\, #f)
          )
        )
      )
    )
    ( (real-list)
      (call-with-current-continuation
        (lambda (return)
          (map
            (lambda (x) 
              (if (not (string->number x))
                (return #f)
              )
              (string->number x)
            )
            (text-split-with-empty string #\, #f)
          )
        )
      )
    )
    ( (*multiple*) '*multiple*)
    (else #f)
  )
)


;; converts the user defined properties (stored with objects) to strings,
;; so they can be displayed in the user interface
(define (object-property->string type property)
  (case type
    ( (word)    property )
    ( (integer) (number->string property))
    ( (real)    (number->string property))
    ( (word-list)
      (let* 
        ( (n (length property))
          (text (car property))
        )
        (do ((i 1 (1+ i))) ((= i n))
          (set! text (string-append text "," (list-ref property i)))
        )
        text
      )
    )
    ( (integer-list)
      (let* 
        ( (n (length property))
          (text (number->string (car property)))
        )
        (do ((i 1 (1+ i))) ((= i n))
          (set! text (string-append text "," 
                                    (number->string (list-ref property i))))
        )
        text
      )
    )
    ( (real-list)
      (let* 
        ( (n (length property))
          (text (number->string (car property)))
        )
        (do ((i 1 (1+ i))) ((= i n))
          (set! text (string-append text "," 
                                    (number->string (list-ref property i))))
        )
        text
      )
    )
    ; if something wrong at least return empty string
    ( else
      ""
    )
  )
)




; --------------------------------------------------------------------
; Functions for UNDO aware property changes
; --------------------------------------------------------------------

(define (object-property-undo-del type id prop_id undo allow-internal)
  (let*
    ( (prop (object-property-get type id prop_id)) )
    (if prop
      (if undo
        (undo-add (list object-property-undo-set 
                        (list 'quote type)
                        (list 'quote id)
                        (list 'quote prop_id)
                        (list 'quote prop)
                        #f
                        allow-internal))
        (redo-add (list object-property-undo-set 
                        (list 'quote type)
                        (list 'quote id)
                        (list 'quote prop_id)
                        (list 'quote prop)
                        #t
                        allow-internal))
      )
    )
  )
  (object-property-del type id prop_id allow-internal)
)


(define (object-property-undo-set type id prop_id prop undo allow-internal)
  (let*
    ( (prop (object-property-get type id prop_id)) )
    ; check whether property exists
    (if prop
      ; we are changing an already existing property
      (if undo
        (undo-add (list object-property-undo-set 
                        (list 'quote type)
                        (list 'quote id)
                        (list 'quote prop_id)
                        (list 'quote prop)
                        #f
                        allow-internal))
        (redo-add (list object-property-undo-set 
                        (list 'quote type)
                        (list 'quote id)
                        (list 'quote prop_id)
                        (list 'quote prop)
                        #t
                        allow-internal))
      )
      ; property does not exist, so the UNDO action is deletion
      (if undo
        (undo-add (list object-property-undo-del 
                        (list 'quote type)
                        (list 'quote id)
                        (list 'quote prop_id)
                        #f
                        allow-internal))
        (redo-add (list object-property-undo-del 
                        (list 'quote type)
                        (list 'quote id)
                        (list 'quote prop_id)
                        #t
                        allow-internal))
      )
    )
  )
  (object-property-set type id prop_id prop allow-internal)
)



(define (object-property-undo-update elem-list new-prop undo allow-internal)
  (for-each
    (lambda (obj)
      ; delete properties which does not exist any more
      (for-each
        (lambda (prop)
          (if (not (assoc (car prop) new-prop))
            (object-property-undo-del (car obj) 
                                      (cadr obj) 
                                      (car prop)
                                      undo
                                      allow-internal)
          )
        )
        (object-property-get-all-user (car obj) (cadr obj))
      )

      ; set the new properties
      (for-each
        (lambda (prop)
          ; if no multiple keyword, then set it
          (if (not (or (equal? (cadr prop) '*multiple*)
                       (equal? (caddr prop) '*multiple*)))
            (begin
              (object-property-undo-set (car obj) 
                                        (cadr obj) 
                                        (car prop)
                                        (cdr prop)
                                        undo
                                        allow-internal)
            )
          )
        )
        new-prop
      )
    )
    elem-list
  )

)

