(define-module (sx render))
(use-modules (open-gl gl) (open-gl glu)
             (sx property) (sx selection-basic) (sx trackball))
(export render-style-get render-style-set
        render-init render-scene)


; Viewport width and height
(define-public *GL_VIEWPORT_WIDTH* #f)
(define-public *GL_VIEWPORT_HEIGHT* #f)

; center of rotation
(define-public *GL_CENTER*      '(0 0 0))

; Viewing parameters
(define-public *QUATERNION* '(0 0 0 1))
(define-public *ZOOM*   1.0)
(define-public *XDIST* 0.0)
(define-public *YDIST* 0.0)

; do we want lighting
(define-public *GL_RENDER_LIGHTING* #t)

; the color of selection
(define-public *GL_COLOR_SELECTION* '(1 0 0 0.5))



; number to use during rendering to identify the object type
(define *OBJECT_POINT_ID*  11)
(define *OBJECT_LINK_ID*   22)
(define *OBJECT_TRIANG_ID* 33)
(define *OBJECT_QUAD_ID*   44)
(define *OBJECT_BOX_ID*    55)
(define *OBJECT_GROUP_ID*  100)


; rendering properties of the different objects
(define *GL_RENDER_OBJECT_STYLE*
  '( (p (color (1 1 0 0.75))
        (show #t)
        (style lines) ; cannot change
     )
     (l (color (1 1 1 0.75))
        (show #t)
        (tiling #f)
        (tile_percent 0.8)
        (style lines) ; cannot change
     )
     (t (color (0 0 1 0.5))
        (show #t)
        (tiling #t)
        (tile_percent 0.8)
        (style opaq)
     )
     (q (color (0 1.0 0 0.5))
        (show #t)
        (tiling #t)
        (tile_percent 0.8)
        (style transparent)
     )
     (b (color (1 1 1 0.5))
        (show #t)
        (tiling #f)
        (tile_percent 0.8)
        (style transparent)
     )
     (g (color (0 1 0 0.5))
        (show #t)
        (tiling #f)
        (tile_percent 0.8)
        (style opaq)
     )
   )
)



;; accessing rendering properties
(define (render-style-get type property)
  (cadr (assoc property (cdr (assoc type *GL_RENDER_OBJECT_STYLE*))))
)


;; setting rendering properties
(define (render-style-set type property value)
  (let*
    ( (new-style (assoc-set! (cdr (assoc type *GL_RENDER_OBJECT_STYLE*))
                             property 
                             (list value)))
    )
    (set! *GL_RENDER_OBJECT_STYLE* (assoc-set! *GL_RENDER_OBJECT_STYLE*
                                               type 
                                               new-style))
  )
)


; This functions renders the element with nodes as a line object.
; If tiling is allowed it reduces the size of the element before rendering.
(define (render-with-style nodes tiling percent lighting)
  ; normal calculation if necessary
  (if lighting
    (let*
      ; c2 is accessed twice
      ( (c2 (object-property-get 'p (cadr nodes) 'coords))
        (t1 (map (lambda (x y) (- x y))
                 (object-property-get 'p (car nodes) 'coords)
                 c2)
        )
        (t2 (map (lambda (x y) (- x y))
                 c2 
                 (object-property-get 'p (caddr nodes) 'coords))
        )
      )
      ; it will be normalised by OpenGL
      (gl-normal3d (- (* (cadr t1) (caddr t2))
                      (* (cadr t2) (caddr t1)))
                   (- (* (car t2)  (caddr t1))
                      (* (car t1)  (caddr t2)))
                   (- (* (car t1)  (cadr t2))
                      (* (car t2)  (cadr t1)))
      )
    )
  )

  ; if the element should be tiled
  (if tiling
    (let* ( (xcenter '(0.0 0.0 0.0)) (n (length nodes)) )
      ; determine the centre of the element
      (for-each 
        (lambda (x)
          (set! xcenter (map (lambda (x y) (+ x y))
                             xcenter 
                             (object-property-get 'p x 'coords) )
          )
        )
        nodes
      )
      (set! xcenter (map (lambda (x) (/ x n)) xcenter))
      ; draw the tiled element
      ; x' = xc + (x - xc) * percent
      (for-each 
        (lambda (x)
          (gl-vertex3dv
            (map (lambda (x y) (+ x y))
                 (map (lambda (x) (* x percent))
                      (map (lambda (x y) (- x y)) 
                           (object-property-get 'p x 'coords)
                           xcenter)
                 )
                 xcenter
            )
          )
        )
        nodes
      )
    )
    ; non-tiled element
    (for-each 
      (lambda (x)
        (gl-vertex3dv (object-property-get 'p x 'coords))
      )
      nodes
    )
  )
)


; --------------------------------------------------------
; Points
; --------------------------------------------------------

; This function renders one point.
(define (point-one-render-only id)
  (let* ( (coords (object-property-get 'p id 'coords)) )
    (gl-begin 'points)
    (gl-vertex3d  (car coords) (cadr coords) (caddr coords))
    (gl-end)
  )
)

; This function renders the points to an OpenGL area.
; mode - rendering mode
(define (points-render-basic mode)
  (if (not (null? *POINTS*))
    (let* 
      ( (color (render-style-get 'p 'color)) )

      (gl-load-name *OBJECT_POINT_ID*) ; point id
      (gl-push-name 0)                 ; point number

      ; set the color of points
      (gl-color4dv color)
      (for-each 
        (lambda (x)
          ; render the point only if it is not in a group
          (if (not (object-property-get 'p (car x) 'group))
            (begin
              (gl-load-name (car x))
              (if (member (list 'p (car x)) *SELECT_LIST*)
                (begin
                  (gl-color4dv *GL_COLOR_SELECTION*)
                  (point-one-render-only (car x))
                  (gl-color4dv color)
                )
                (point-one-render-only (car x))
              )
            )
          )
        )
        *POINTS*
      )

      (gl-pop-name)
    )
  )
)

; --------------------------------------------------------
; Links
; --------------------------------------------------------

(define (link-one-render-only id tiling percent)
  (let*
    ( (nodes (object-property-get 'l id 'nodes)) )

    (gl-begin 'lines)
    (render-with-style nodes tiling percent #f)
    (gl-end)

  )
)


(define (links-render-basic mode)
  (if (not (null? *LINKS*))
    (let* 
      ( (color (render-style-get 'l 'color))
        (tiling (render-style-get 'l 'tiling))
        (percent (render-style-get 'l 'tile_percent))
      )

      (gl-load-name *OBJECT_LINK_ID*) ; link id
      (gl-push-name 0)                ; link number

      (gl-color4dv color)
      (for-each 
        (lambda (x)
          ; render the link only if it is not in a group
          (if (not (object-property-get 'l (car x) 'group))
            (begin
              (gl-load-name (car x))
              (if (member (list 'l (car x)) *SELECT_LIST*)
                (begin
                  (gl-color4dv *GL_COLOR_SELECTION*)
                  (link-one-render-only (car x) tiling percent)
                  (gl-color4dv color)
                )
                (link-one-render-only (car x) tiling percent)
              )
            )
          )
        )
        *LINKS*
      )

      (gl-pop-name)
    )
  )
)

; --------------------------------------------------------
; Triangles
; --------------------------------------------------------

(define (triang-one-render-only id style tiling percent)
  (let*
    ( (nodes (object-property-get 't id 'nodes)) )

    (if (eq? style 'lines)
      (gl-begin 'line-loop)
      (gl-begin 'triangles)
    )
    (render-with-style nodes tiling percent *GL_RENDER_LIGHTING*)
    (gl-end)
  )
)


(define (triangs-render-basic mode)
  (if (not (null? *TRIANGS*))
    (let* 
      ( (color (render-style-get 't 'color))
        (tiling (render-style-get 't 'tiling))
        (percent (render-style-get 't 'tile_percent))
        (style (render-style-get 't 'style))
      )

      (gl-load-name *OBJECT_TRIANG_ID*) ; triang id
      (gl-push-name 0)                  ; triang number

      (gl-color4dv color)
      (for-each 
        (lambda (x)
          ; render the triangle only if it is not in a group
          (if (not (object-property-get 't (car x) 'group))
            (begin
              (gl-load-name (car x))
              (if (member (list 't (car x)) *SELECT_LIST*)
                (begin
                  (gl-color4dv *GL_COLOR_SELECTION*)
                  (triang-one-render-only (car x) style tiling percent)
                  (gl-color4dv color)
                )
                (triang-one-render-only (car x) style tiling percent)
              )
            )
          )
        )
        *TRIANGS*
      )

      (gl-pop-name)
    )
  )
)

; --------------------------------------------------------
; Quads
; --------------------------------------------------------

(define (quad-one-render-only id style tiling percent)
  (let*
    ( (nodes (object-property-get 'q id 'nodes)) )

    (if (eq? style 'lines)
      (gl-begin 'line-loop)
      (gl-begin 'quads)
    )
    (render-with-style nodes tiling percent *GL_RENDER_LIGHTING*)
    (gl-end)
  )
)


(define (quads-render-basic mode)
  (if (not (null? *QUADS*))
    (let* 
      ( (color (render-style-get 'q 'color))
        (tiling (render-style-get 'q 'tiling))
        (percent (render-style-get 'q 'tile_percent))
        (style (render-style-get 'q 'style))
      )

      (gl-load-name *OBJECT_QUAD_ID*) ; quad id
      (gl-push-name 0)                ; quad number

      (gl-color4dv color)
      (for-each 
        (lambda (x)
          ; render the quad only if it is not in a group
          (if (not (object-property-get 'q (car x) 'group))
            (begin
              (gl-load-name (car x))
              (if (member (list 'q (car x)) *SELECT_LIST*)
                (begin
                  (gl-color4dv *GL_COLOR_SELECTION*)
                  (quad-one-render-only (car x) style tiling percent)
                  (gl-color4dv color)
                )
                (quad-one-render-only (car x) style tiling percent)
              )
            )
          )
        )
        *QUADS*
      )

      (gl-pop-name)
    )
  )
)


; --------------------------------------------------------
; Boxes
; --------------------------------------------------------

(define (box-one-render-only id style tiling percent)
  (let*
    ( (nodes (object-property-get 'b id 'nodes))
      (coord1 (object-property-get 'p (car nodes) 'coords))
      (coord2 (object-property-get 'p (cadr nodes) 'coords))
      (xmin #f) (xmax #f) (ymin #f) (ymax #f) (zmin #f) (zmax #f)
    )
    (if (< (car coord1) (car coord2))
      (begin 
        (set! xmin (car coord1))
        (set! xmax (car coord2))
      )
      (begin
        (set! xmin (car coord2))
        (set! xmax (car coord1))
      )
    )
    (if (< (cadr coord1) (cadr coord2))
      (begin 
        (set! ymin (cadr coord1))
        (set! ymax (cadr coord2))
      )
      (begin
        (set! ymin (cadr coord2))
        (set! ymax (cadr coord1))
      )
    )
    (if (< (caddr coord1) (caddr coord2))
      (begin 
        (set! zmin (caddr coord1))
        (set! zmax (caddr coord2))
      )
      (begin
        (set! zmin (caddr coord2))
        (set! zmax (caddr coord1))
      )
    )

    (gl-begin 'quads)
    (for-each
      (lambda (faces)
        (if *GL_RENDER_LIGHTING*
          (gl-normal3dv (car faces))
        )
        (for-each (lambda (coord) (gl-vertex3dv coord)) (cdr faces))
      )
      (list (list '(0 0 1)
                  (list xmin ymin zmin) (list xmax ymin zmin)
                  (list xmax ymax zmin) (list xmin ymax zmin) )
            (list '(1 0 0)
                  (list xmin ymin zmin) (list xmin ymax zmin)
                  (list xmin ymax zmax) (list xmin ymin zmax) )
            (list '(0 1 0)
                  (list xmax ymax zmin) (list xmin ymax zmin)
                  (list xmin ymax zmax) (list xmax ymax zmax) )
            (list '(1 0 0)
                  (list xmax ymin zmin) (list xmax ymax zmin)
                  (list xmax ymax zmax) (list xmax ymin zmax) )
            (list '(0 1 0)
                  (list xmax ymin zmin) (list xmin ymin zmin)
                  (list xmin ymin zmax) (list xmax ymin zmax) )
            (list '(0 0 1)
                  (list xmin ymin zmax) (list xmax ymin zmax)
                  (list xmax ymax zmax) (list xmin ymax zmax) )
      )
    )
    (gl-end)
  )
)

(define (boxes-render-basic mode)
  (if (not (null? *BOXES*))
    (let* 
      ( (color (render-style-get 'q 'color))
        (tiling #f)
        (percent #f)
        (style (render-style-get 'q 'style))
      )

      (gl-load-name *OBJECT_BOX_ID*)
      (gl-push-name 0)

      (gl-color4dv color)
      (for-each 
        (lambda (x)
          ; render the box only if it is not in a group
          (if (not (object-property-get 'b (car x) 'group))
            (begin
              (gl-load-name (car x))
              (if (member (list 'b (car x)) *SELECT_LIST*)
                (begin
                  (gl-color4dv *GL_COLOR_SELECTION*)
                  (box-one-render-only (car x) style tiling percent)
                  (gl-color4dv color)
                )
                (box-one-render-only (car x) style tiling percent)
              )
            )
          )
        )
        *BOXES*
      )

      (gl-pop-name)

    )
  )
)

; --------------------------------------------------------
; Groups
; --------------------------------------------------------

(define (group-one-render-only id tiling percent)
  (let* 
    ( (parts (object-property-get 'g id 'parts)) 
      (style (render-style-get 'g 'style))
    )
    ; first surface elements
    (for-each 
      (lambda (y)
        (case (car y)
          ( (t) (triang-one-render-only (cadr y) style tiling percent) )
          ( (q) (quad-one-render-only   (cadr y) style tiling percent) )
          ( (b) (box-one-render-only    (cadr y) style tiling percent) )
          ( (g) (group-one-render-only  (cadr y) tiling percent) )
        )
      )
      parts
    )
    ; switch off lighting
    (if *GL_RENDER_LIGHTING*
      (gl-disable 'lighting)
    )
    (for-each 
      (lambda (y)
        (case (car y)
          ( (p) (point-one-render-only  (cadr y)) )
          ( (l) (link-one-render-only   (cadr y) tiling percent) )
        )
      )
      parts
    )
    ; enable lighting if necessary
    (if *GL_RENDER_LIGHTING*
      (gl-enable 'lighting)
    )
  )
)


(define (groups-render-basic mode)
  (if (not (null? *GROUPS*))
    (let* 
      ( (color (render-style-get 'g 'color))
        (tiling (render-style-get 'g 'tiling))
        (percent (render-style-get 'g 'tile_percent))
      )

      (gl-load-name *OBJECT_GROUP_ID*)
      (gl-push-name 0)

      (gl-color4dv color)
      (for-each 
        (lambda (x)
          (if (not (object-property-get 'g (car x) 'group))
            (begin
              (gl-load-name (car x))
              (if (member (list 'g (car x)) *SELECT_LIST*)
                (begin
                  (gl-color4dv *GL_COLOR_SELECTION*)
                  (group-one-render-only (car x) tiling percent)
                  (gl-color4dv color)
                )
                (group-one-render-only (car x) tiling percent)
              )
            )
          )
        )
        *GROUPS*
      )
      (gl-pop-name)
    )
  )
)

; --------------------------------------------------------
; Scene
; --------------------------------------------------------

(define (render-all-objects req-style mode)
  (for-each
    (lambda (obj-type)
      (let* 
        ( (show (render-style-get obj-type 'show))
          (style (render-style-get obj-type 'style))
        )
        (if (and show (eq? style req-style))
          (case obj-type
            ( (p) (points-render-basic mode) )
            ( (l) (links-render-basic mode) )
            ( (t) (triangs-render-basic mode) )
            ( (q) (quads-render-basic mode) )
            ( (b) (boxes-render-basic mode) )
            ( (g) (groups-render-basic mode) )
          )
        )
      )
    )
    '(p l t q b g)
  )
)


; initializes the OpenGL parameters
(define (render-init)
  ; enable hidden line
  (gl-enable 'depth-test)
  ; polygon rendering mode settings
  (gl-shade-model 'smooth)
  (gl-hint 'polygon-smooth 'fastest)
  (gl-polygon-mode 'front-and-back 'fill)
  ; point size
  (gl-point-size 7)
  ; clear color for the background
  (gl-clear-color 0 0 0 1)
  ; enable colored material (?)
  (gl-enable 'color-material)
  (gl-color-material 'front-and-back 'ambient-and-diffuse)

  ; enable lighting
  ; ---------------
  (gl-enable 'lighting)
  ; enable light 0
  (gl-enable 'light0)
  ; both sides of the polygon should be shown
  (gl-light-modelf 'light-model-two-side 1)
  ; no local view, faster rendering
  (gl-light-modelf 'light-model-local-viewer 0)

  ; lighting conditions, diffusion is more important
  (gl-lightfv 'light0 'ambient  '(0.1 0.1 0.1 1.0))
  (gl-lightfv 'light0 'diffuse  '(0.7 0.7 0.7 0.0))
  ; if I want some shining, I would need this
;  (gl-lightfv 'light0 'specular '(0.0 0.0 0.0 0.0))

  ; material properties to control, not necessary
;  (gl-materialfv 'front-and-back 'ambient '(0.0 0.0 0.0 1.0))
;  (gl-materialfv 'front-and-back 'shininess '(40.0))
;  (gl-materialfv 'front-and-back 'specular '(0.0 0.0 0.0 0.0))
;  (gl-materialfv 'front-and-back 'diffuse '(1.0 0.0 0.0 0.0))

  ; global ambient light, not necessary
  (gl-light-modelfv 'light-model-ambient '(0.2 0.2 0.2 1.0))

  ; this is very important !!!
  ; during zooming the transformation applies to the normals as well
  ; so either I consider that explicitly in the normal calculation
  ; or I let OpenGL to do it
  (gl-enable 'normalize)
  ; position of light, default
  ; only direction is given, no position
  (gl-lightfv 'light0 'position (list 0.0 0.0 1.0 0.0))

  ; store the rendering of the orientation
  (gl-new-list 101 'compile)
  (gl-color3d 1.0 1.0 1.0)
  (gl-begin 'lines)
  ; x axis
  (gl-vertex3d 0.0 0.0 0.0)
  (gl-vertex3d 1.0 0.0 0.0)
  ; y axis
  (gl-vertex3d 0.0 0.0 0.0)
  (gl-vertex3d 0.0 1.0 0.0)
  ;z axis
  (gl-vertex3d 0.0 0.0 0.0)
  (gl-vertex3d 0.0 0.0 1.0)

  ; x text
  (gl-vertex3d 0.8 0.1 0.0)
  (gl-vertex3d 1.0 0.3 0.0)
  (gl-vertex3d 1.0 0.1 0.0)
  (gl-vertex3d 0.8 0.3 0.0)
  ; y text
  (gl-vertex3d 0.2 0.8 0.0)
  (gl-vertex3d 0.2 0.9 0.0)
  (gl-vertex3d 0.2 0.9 0.0)
  (gl-vertex3d 0.1 1.0 0.0)
  (gl-vertex3d 0.2 0.9 0.0)
  (gl-vertex3d 0.3 1.0 0.0)
  ;z text
  (gl-vertex3d 0.1 0.1 1.0)
  (gl-vertex3d 0.3 0.1 1.0)
  (gl-vertex3d 0.1 0.1 1.0)
  (gl-vertex3d 0.3 0.3 1.0)
  (gl-vertex3d 0.3 0.3 1.0)
  (gl-vertex3d 0.1 0.3 1.0)

  (gl-end)
  (gl-end-list)
)



; This function renders the 3D scene
; returns #f or the list of selected objects
(define (render-scene mode x y pick_x pick_y compile)
  (let*
    ( (max 1.0) )
    (if (eq? mode 'select)
      (begin
        (gl-select-buffer *GL_SELECT_BUFFER*)
        (gl-render-mode 'select)
      )
    )
    ; names are always generated
    ; so the selection will be faster
    (gl-init-names)
    (gl-push-name 0)

    ; clear buffers
    (gl-clear '(color-buffer depth-buffer))

    ; first projection
    (gl-matrix-mode 'projection)
    (gl-load-identity)
    ; when selecting make picking projections
    (if (eq? mode 'select)
      (begin
        ; it should be protected when the width or height is zero
        ; at least 1 pixel width or height is specified
        ; otherwise it would select absolutely everything !!!
        (if (= pick_x 0) (set! pick_x 1))
        (if (= pick_y 0) (set! pick_y 1))
        (glu-pick-matrix  x (- *GL_VIEWPORT_HEIGHT* y)
                          pick_x pick_y
                          (list 0 0 *GL_VIEWPORT_WIDTH* *GL_VIEWPORT_HEIGHT*)
        )
      )
    )
    ; normal projection
    ; FIXME: I am still not sure about this, 
    ; Is it big enough not to be clipped ?
    (set! max (* *ZOOM* *POINT_MAX* 2.0))
    (if (< max 1.0) (set! max 1.0))
    (gl-ortho -1 1 -1 1 (- max) max)

    ; then model view handling
    (gl-matrix-mode 'modelview)
    (gl-load-identity)

    (gl-scaled *ZOOM* *ZOOM* *ZOOM*)
    (gl-translated *XDIST* *YDIST* 0)
    ; rotation around origo
    (gl-mult-matrixd (trackball-rotmatrix *QUATERNION*))
    ; displacement of the origo to other position
    (gl-translated (- (car *GL_CENTER*))
                   (- (cadr *GL_CENTER*))
                   (- (caddr *GL_CENTER*)) )

    ; enable depth, it is switched off later
    (gl-enable 'depth-test)
    ; then lights are on or off
    (if *GL_RENDER_LIGHTING*
      (gl-enable 'lighting)
      (gl-disable 'lighting)
    )

    (if compile
      (begin
        (gl-disable 'blend)
        (gl-depth-mask #t)
        (gl-new-list 10 'compile-and-execute)
        ; first render opaq objects
        (render-all-objects 'opaq mode)
        (gl-end-list)

        ; then render transparent objects
        ; for transparent look 
        ; switch on blending, specify blending function and
        ; disable depth buffer (read-only mode), 
        ; this latter is necessary so
        ; it is always transparent, and 
        ; the rendering does not depend on the element order
        (gl-enable 'blend)
        (gl-blend-func 'src-alpha 'dst-alpha)
        (gl-depth-mask #f)
        (gl-new-list 11 'compile-and-execute)
        (render-all-objects 'transparent mode)
        (gl-end-list)

        ; switch off any transparency
        (gl-disable 'blend)
        (gl-depth-mask #t)

        (gl-disable 'lighting)
        ; non surface objects are rendered here without lights
        (gl-new-list 12 'compile-and-execute)
        (render-all-objects 'lines mode)
        (gl-end-list)

      )
      (begin
        (gl-disable 'blend)
        (gl-depth-mask #t)
        (gl-call-list 10)

        (gl-enable 'blend)
        (gl-blend-func 'src-alpha 'dst-alpha)
        (gl-depth-mask #f)
        (gl-call-list 11)

        (gl-disable 'blend)
        (gl-depth-mask #t)

        (gl-disable 'lighting)
        (gl-call-list 12)
      )
    )

    ; clean up name stack
    (gl-pop-name)
    ; postprocessing of the screen
    (case mode 
      ; selection
      ( (select)
        (let*
          ; switch back to rendering, get number of selections
          ((n (gl-render-mode 'render))
           (type #f) (id   #f) (out  '())
          )

          (do ((i 0 (1+ i))) ((= i n))
            (set! type (array-ref *GL_SELECT_BUFFER* (+ (* i 5) 3)))
            (set! id   (array-ref *GL_SELECT_BUFFER* (+ (* i 5) 4)))
            ; transform it
            (cond 
              ( (= type *OBJECT_POINT_ID*)
                (set! out (cons (list 'p id) out)) )
              ( (= type *OBJECT_LINK_ID*)
                (set! out (cons (list 'l id) out)) )
              ( (= type *OBJECT_TRIANG_ID*)
                (set! out (cons (list 't id) out)) )
              ( (= type *OBJECT_QUAD_ID*)
                (set! out (cons (list 'q id) out)) )
              ( (= type *OBJECT_BOX_ID*)
                (set! out (cons (list 'b id) out)) )
              ( (= type *OBJECT_GROUP_ID*)
                (set! out (cons (list 'g id) out)) )
            )
          )
          (if (null? out)
            #f
            out
          )
        )
      )
      ( (render)
        ; no depth rendering
        (gl-disable 'depth-test)
        ; no lighting
        (gl-disable 'lighting)
        ; render orientation at the bottom left corner
        ; save the current viewport
        (gl-push-attrib '(viewport))
        ; top left corner to render into
        (gl-viewport 0 0 100 100)
        (gl-matrix-mode 'modelview)
        (gl-load-identity)

        ; rotation around origo
        (gl-mult-matrixd (trackball-rotmatrix *QUATERNION*))
        ; render the orientation
        (gl-call-list 101)
        ; reset the viewport back to original
        (gl-pop-attrib)

        ; FIXME: this is actually PICK_MODE dependent
        ; select_cross -> window draw,
        ; select_obj -> picking cursor draw
        ; select_coord & SELECT_BASE -> line from BASE to cursor
        (if *SELECT_WIN_CORNER_2*
          (begin
            (gl-matrix-mode 'projection)
            (gl-load-identity)
            (glu-ortho2d 0 *GL_VIEWPORT_WIDTH* 0 *GL_VIEWPORT_HEIGHT*)
            (gl-matrix-mode 'modelview)
            (gl-load-identity)
            (gl-color3d 1 1 1)
            (gl-begin 'line-loop)
              (gl-vertex2d (car *SELECT_WIN_CORNER_1*)
                           (- *GL_VIEWPORT_HEIGHT* (cadr *SELECT_WIN_CORNER_1*)))
              (gl-vertex2d (car *SELECT_WIN_CORNER_2*) 
                           (- *GL_VIEWPORT_HEIGHT* (cadr *SELECT_WIN_CORNER_1*)))
              (gl-vertex2d (car *SELECT_WIN_CORNER_2*) 
                           (- *GL_VIEWPORT_HEIGHT* (cadr *SELECT_WIN_CORNER_2*)))
              (gl-vertex2d (car *SELECT_WIN_CORNER_1*) 
                           (- *GL_VIEWPORT_HEIGHT* (cadr *SELECT_WIN_CORNER_2*)))
            (gl-end)
          )
        )
        ; return value
        #f
      )
    )
  )
)





