(define-module (sx selection))
(use-modules (sx property) (sx selection-basic) (sx basic) (sx view)
             (sx gui-entry) (sx gui-gl) (sx gui-selection))
(export selection-add-ref-nodes selection-add-internal
        selection-object selection-coord selection-dist-point 
        selection-cancel)


; this function adds all those points to the selection which are
; referenced by a selected element
(define (selection-add-ref-nodes expand)
  (let*
    ( (nodes #f) )
    ; add nodes of elements, no duplicates, no inverting selection
    (set! *SELECT_OBJ_MODE* 'add)
    (for-each 
      (lambda (x)
        (case (car x)
          ( (g) 
            (if expand 
              (set! nodes (group-get-points-basic (cadr x)))
              (set! nodes #f)
            )
          )
          ( (l t q b)
            (set! nodes (map (lambda (y) (list 'p y))
                             (object-property-get (car x) (cadr x) 'nodes)))
          )
        )
        (if nodes
          (selection-add-internal nodes)
        )
      )
      *SELECT_LIST*
    )
  )
)

; adds/removes the object to/from the selection list
; sellist  - the list of selections
; it returns the number of duplicates
(define (selection-add-internal sellist)
  (let* ((dup 0))
    (for-each 
      (lambda (sel)
        (if (member  sel *SELECT_LIST*) ; if in the list
          (begin
            (case *SELECT_OBJ_MODE*
              ( (invert remove)
                (set! *SELECT_LIST* (delete! sel *SELECT_LIST*))
              )
            )
            (set! dup (1+ dup))
          )
          (case *SELECT_OBJ_MODE*
            ( (add invert)
              (set! *SELECT_LIST* (cons sel *SELECT_LIST*))
            )
          )
        )
      )
      sellist
    )
    dup
  )
)


; These are the internal keywords for object selection,
; external keywords (other than these) can be defined through select-init.
; The selection-object-internal function uses this list to validate its
; internal keywords !
(define *KEYWORDS_OBJECT_SELECTION* 
  '("all" "last" "previous" "crossing"
    "add" "remove" "invert"
    "point" "link" "triangle" "quad" "box" "group"
    "only_point" "only_link" "only_triang" "only_quad" "only_box"
    "only_group" "any_object"
    "single" "multiple"
    "view")
)


(define (selection-object-prompt-update)
  (let* ( (prompt "") )
    ; notify, that only certain element types can be selected
    (case *SELECT_USER_TYPE*
      ( (p)  (set! prompt "Select ONLY point(s)") )
      ( (l)  (set! prompt "Select ONLY link(s)") )
      ( (t)  (set! prompt "Select ONLY triang(s)") )
      ( (q)  (set! prompt "Select ONLY quad(s)") )
      ( (b)  (set! prompt "Select ONLY box(es)") )
      ( (g)  (set! prompt "Select ONLY group(s)") )
      ( else (set! prompt *SELECT_TEXT_PREFIX*) )
    )
    (set! prompt (string-append prompt " <"))
    ; notify, that only single selection
    (if *SELECT_USER_SINGLE*
      (begin
        (set! prompt (string-append prompt "single/"))
      )
    )
    ; write back the current prefix
    (case *SELECT_PREFIX*
      ( (p) (set! prompt (string-append prompt "point> ")) )
      ( (l) (set! prompt (string-append prompt "link> ")) )
      ( (t) (set! prompt (string-append prompt "triang> ")) )
      ( (q) (set! prompt (string-append prompt "quad> ")) )
      ( (b) (set! prompt (string-append prompt "box> ")) )
      ( (g) (set! prompt (string-append prompt "group> ")) )
      ( else (set! prompt (string-append prompt "*unknown*> ")) )
    )
    (gui-entry-label-set-text prompt)
  )
)


; returns the list of objects in *SELECT_LIST*
(define (selection-object-internal selection action)
  (let* 
    ( (remove-crossing (if (eq? *PICK_MODE* 'select_cross)
                             (begin
                               (selection-crossing-quit)
                               #t
                             )
                             #f))
    )
    ; interpret the selected object list and keywords
    (case (length selection)
      ; no data on an ENTER, so selection is presumably finished
      ( (0)
        (if remove-crossing
          (gui-update-gl-window #f)
        )
        ; check whether empty selection is allowed
        (if (and (member 'no_empty *SELECT_INIT*) (null? *SELECT_LIST*))
          (begin
            (gui-entry-add-history " *No object is selected*")
            (set! *SELECT_STATE* 'init)
          )
          (if (null? *SELECT_LIST*)
            ; when empty selection, signal as cancel
            (begin
              (set! *SELECT_STATE* 'cancel)
              (action)
            )
            ; signal, that there are selected objects
            (begin
              (set! *SELECT_STATE* 'objects)
              (action)
            )
          )
        )
      )
      ; there is something in the new selection list, so lets process them
      (else 
        (let*
          ( (after_mesg "")         ; messages to print after statistics
            (dup 0)                 ; number of duplicate selections
            (objlist '())           ; list to collect valid objects
            (prompt #t)             ; whether to update the prompt
          )
          (call-with-current-continuation (lambda (return)
          ; go through the new selection
          (for-each
            (lambda (x)
              (if (number? x) ; if it is a number
                ; according to the prefix
                (case *SELECT_PREFIX*
                  ( (p l t q b g)
                    ; validate the object and store it, 
                    (if (object-valid (list *SELECT_PREFIX* x))
                      (set! objlist (cons (list *SELECT_PREFIX* x) objlist))
                    )
                  )
                  ; notify the user, that something is wrong
                  (else
                    (gui-entry-add-history "*Internal error* - unknown prefix")
                    (return #f)
                  )
                )
                ; is it an internal keyword ?
                (let* ((words '()))
                  ; collect matching internal keywords
                  (for-each (lambda (y)
                    (if (and (<= (string-length x) (string-length y))
                             (string-ci=? x (substring y 0 (string-length x))) )
                      (set! words (cons y words))
                    ) )
                    *KEYWORDS_OBJECT_SELECTION*
                  )
                  ; determine what to do
                  (case (length words)
                    ; only one keyword is found
                    ( (1)
                      ; prefix is changing or some other selection mechanism
                      (case (string->symbol (car words))
                        ; what kind of element are we selecting presently
                        ( (point)    (set! *SELECT_PREFIX* 'p))
                        ( (link)     (set! *SELECT_PREFIX* 'l))
                        ( (triangle) (set! *SELECT_PREFIX* 't))
                        ( (quad)     (set! *SELECT_PREFIX* 'q))
                        ( (box)      (set! *SELECT_PREFIX* 'b))
                        ( (group)    (set! *SELECT_PREFIX* 'g))

                        ; different modes of object selection
                        ( (add)    (set! *SELECT_OBJ_MODE* 'add) )
                        ( (remove) (set! *SELECT_OBJ_MODE* 'remove) )
                        ( (invert) (set! *SELECT_OBJ_MODE* 'invert) )

                        ; only certain element types can be selected
                        ; this restriction is by the user
                        ( (only_point)
                          (if (or (member 'type_any *SELECT_INIT*)
                                  (member 'type_point *SELECT_INIT*) )
                            (set! *SELECT_USER_TYPE* 'p)
                          )
                        )
                        ( (only_link)
                          (if (or (member 'type_any *SELECT_INIT*)
                                  (member 'type_link *SELECT_INIT*) )
                            (set! *SELECT_USER_TYPE* 'l)
                          )
                        )
                        ( (only_triang)
                          (if (or (member 'type_any *SELECT_INIT*)
                                  (member 'type_triang *SELECT_INIT*) )
                            (set! *SELECT_USER_TYPE* 't)
                          )
                        )
                        ( (only_quad)
                          (if (or (member 'type_any *SELECT_INIT*)
                                  (member 'type_quad *SELECT_INIT*) )
                            (set! *SELECT_USER_TYPE* 'q)
                          )
                        )
                        ( (only_box)
                          (if (or (member 'type_any *SELECT_INIT*)
                                  (member 'type_box *SELECT_INIT*) )
                            (set! *SELECT_USER_TYPE* 'b)
                          )
                        )
                        ( (only_group)
                          (if (or (member 'type_any *SELECT_INIT*)
                                  (member 'type_group *SELECT_INIT*) )
                            (set! *SELECT_USER_TYPE* 'g)
                          )
                        )
                        ; any object is allowed by the user
                        ( (any_object) (set! *SELECT_USER_TYPE* 'any) )

                        ; only one element can be selected
                        ( (single)   (set! *SELECT_USER_SINGLE* #t))
                        ( (multiple) (set! *SELECT_USER_SINGLE* #f))

                        ; we are selecting the last created element
                        ( (last)
                          ; check for valid last created object
                          (if (and *OBJECT_LAST* 
                                   (object-valid *OBJECT_LAST*))
                            (set! objlist (cons *OBJECT_LAST* objlist))
                            (set! after_mesg 
                                  (string-append after_mesg 
                                                 "\n*No valid last object*"))
                          )
                        )
                        ; we are taking the previous selection
                        ( (previous)
                          (let* ( (valid_list '()) )
                            ; validate all previously selected elements
                            (for-each (lambda (x)
                              (if (object-valid x)
                                  (set! valid_list (cons x valid_list))
                              ) )
                              *SELECT_PREVIOUS*
                            )
                            (if (not (null? valid_list))
                              (set! objlist (append valid_list objlist))
                              (set! after_mesg
                                    (string-append after_mesg 
                                                   "\n*No valid previous selection*"))
                            )
                          )
                        )
                        ; we are selecting all of the elements
                        ( (all)
                          (for-each 
                            (lambda (x y)
                              (set! objlist 
                                (append (map (lambda(xx) (list x (car xx))) 
                                             (primitive-eval y))
                                        objlist)
                              )
                            )
                            '(p l t q b g)
                            '(*POINTS* *LINKS* *TRIANGS* *QUADS* *BOXES* *GROUPS*)
                          )
                        )
                        ; switching to multi selection by window
                        ( (crossing)
                          (selection-crossing-enter "First corner: " #f)
                          (set! prompt #f)
                          (return #f)
                        )

                        ; switch to viewing mode
                        ( (view)
                          (if (null? *SELECT_STACK*)
                            ; we want to remember the selected elements
                            (let* ((sel-list *SELECT_LIST*))
                              (selection-push-internal)
                              ; setup viewing
                              (view-setup)
                              ; restore selected elements
                              (set! *SELECT_LIST* sel-list)
                              ; generate new screen
                              (gui-update-gl-window #f)
                              (set! prompt #f)
                            )
                            (set! after_mesg "\nNo reentry to viewing ")
                          )
                          (return #f)
                        )
                      )
                    )
                    ; no internal keyword matches
                    ( (0)
                      (if (null? *SELECT_KEYWORDS*)
                        (set! after_mesg (string-append "\nInvalid entry was " x))
                        ; check for externally defined keywords
                        (selection-keyword-internal x action)
                      )
                    )
                    ; several matching internal keywords
                    (else
                      (set! after_mesg 
                            (string-append "\nAmbiguous keyword was " x " for ["))
                      (for-each (lambda (xx) 
                        (set! after_mesg (string-append after_mesg xx " ")) )
                        words
                      )
                      (set! after_mesg (string-append after_mesg "]"))
                      ; in the case of ambiguous keywords do not process further
                      (return #f)
                    )
                  )
                )
              )
            )
            selection
          )
          )) ; closing the continuation

          ; check whether all object types can be selected, or only certain types
          ; this is required by the command itself, 
          ; it has priority over user restriction
          (if (not (member 'type_any *SELECT_INIT*))
            ; select the allowed object types
            (let* ( (new_list '()) )
              (for-each (lambda (xx)
                  (if (or (and (equal? (car xx) 'p)
                               (member 'type_point *SELECT_INIT*))
                          (and (equal? (car xx) 'l)
                               (member 'type_link *SELECT_INIT*))
                          (and (equal? (car xx) 't)
                               (member 'type_triang *SELECT_INIT*))
                          (and (equal? (car xx) 'q)
                               (member 'type_quad *SELECT_INIT*))
                          (and (equal? (car xx) 'b)
                               (member 'type_box *SELECT_INIT*))
                          (and (equal? (car xx) 'g)
                               (member 'type_group *SELECT_INIT*)) )
                    (set! new_list (cons xx new_list))
                  )
                )
                objlist
              )
              (set! objlist new_list)
            )
          )

          ; user's restriction on the selectable objects
          (if (not (equal? 'any *SELECT_USER_TYPE*))
            ; select the allowed object types
            (let* ( (new_list '()) )
              (for-each (lambda (xx)
;                  (if (or (and (equal? (car xx) 'p)
;                               (equal? 'type_point *SELECT_USER_TYPE*))
;                          (and (equal? (car xx) 'l)
;                               (equal? 'type_link *SELECT_USER_TYPE*))
;                          (and (equal? (car xx) 't)
;                               (equal? 'type_triang *SELECT_USER_TYPE*))
;                          (and (equal? (car xx) 'q)
;                               (equal? 'type_quad *SELECT_USER_TYPE*))
;                          (and (equal? (car xx) 'b)
;                               (equal? 'type_box *SELECT_USER_TYPE*))
;                          (and (equal? (car xx) 'g)
;                               (equal? 'type_group *SELECT_USER_TYPE*)) )
                  (if (eq? (car xx) *SELECT_USER_TYPE*)
                    (set! new_list (cons xx new_list))
                  )
                )
                objlist
              )
              (set! objlist new_list)
            )
          )

          ; do we need only one element per selection ?
          ; this is independent whether we are selecting only one
          ; element type or different elements can be selected
          (if (or (member 'no_multiple *SELECT_INIT*)
                  *SELECT_USER_SINGLE*)
            (begin
              (set! objlist (gui-selection-single-object objlist))
            )
          )

          ; add the valid object to the selection
          (set! dup (selection-add-internal objlist))

          ; update the string to guide selection
          ; valid selection
          (gui-entry-add-history 
            (string-append (number->string (length objlist)) " found"))
          ; notify user about duplicate selection
          (if (> dup 0)
            (gui-entry-add-history 
              (string-append " ( " (number->string dup) " duplicate(s) )"))
          )
          ; total selection, 
          ; Note: all elements in *SELECT_LIST* are valid !!!
          (gui-entry-add-history 
            (string-append ", " 
                           (number->string (length *SELECT_LIST*)) 
                           " total")
          )
          ; there was a change in the selection, update the screen
          (if (null? objlist)
            (if remove-crossing
              (gui-update-gl-window #f)
            )
            (gui-update-gl-window #t)
          )

          ; messages after reporting the found elements
          ; for example report to switch to crossing selection
          (if (> (string-length after_mesg) 0)
            (gui-entry-add-history after_mesg)
          )

          ; update the prompt
          (if prompt
            (selection-object-prompt-update)
          )

          (set! *SELECT_STATE* 'init)
        )
      )
    )
  )
)

(define *KEYWORDS_POINT_SELECTION*
  '("view")
)

; returns the coordinate in *SELECT_X* *SELECT_Y* *SELECT_Z*
(define (selection-coord-internal selection action)
  ; empty entry, cancel operation !!!
  (if (= (length selection) 0)
    (begin
      (if (member 'no_empty *SELECT_INIT*)
        (begin
          (gui-entry-add-history " *Coordinates are required*")
          (set! *SELECT_STATE* 'init)
        )
        (begin
          ; FIXME: this should not be here, the 'cancel state should be enough
          (set! *SELECT_X* #f)
          (set! *SELECT_Y* #f)
          (set! *SELECT_Z* #f)
          (set! *SELECT_STATE* 'cancel)
          ; lets reset everything
          (action)
        )
      )
    )
    (begin
      ; first, allow partial entry or in case of empty entry execute
      (case (length selection)
        ((1) (if (number? (car selection))
               (set! selection (cons ".xyz" (append selection '(0 0))))
             )
        )
        ((2) (if (number? (car selection))
               (set! selection (cons ".xyz" (append selection '(0))))
               (set! selection (append selection '(0 0)))
             )
        )
        ((3) (if (number? (car selection))
               (set! selection (cons ".xyz" selection))
               (set! selection (append selection '(0)))
             )
        )
      )
      ; then validate
      (if (and (= (length selection) 4)
               (string? (car selection))
               (member (string->symbol (car selection))
                       '(.xyz .x .y .z .xy .yx .xz .zx .yz .zy))
               (number? (cadr selection))
               (number? (caddr selection))
               (number? (cadddr selection)) )
        (begin
          ; which components to set
          (case (string->symbol (car selection))
            ((.xyz) (set! *SELECT_X* (cadr selection))
                    (set! *SELECT_Y* (caddr selection))
                    (set! *SELECT_Z* (cadddr selection)) )

            ((.x)   (set! *SELECT_X* (cadr selection)) )
            ((.y)   (set! *SELECT_Y* (caddr selection)) )
            ((.z)   (set! *SELECT_Z* (cadddr selection)) )

            ((.xy .yx)
                    (set! *SELECT_X* (cadr selection))
                    (set! *SELECT_Y* (caddr selection)) )
            ((.xz .zx)
                    (set! *SELECT_X* (cadr selection))
                    (set! *SELECT_Z* (cadddr selection)) )

            ((.yz .zy)
                    (set! *SELECT_Y* (caddr selection))
                    (set! *SELECT_Z* (cadddr selection)) )
          )

          ; write back the users selection to the history
          ; write back the modifier, if any
          (if (not (eq? (string->symbol (car selection)) '.xyz))
            (gui-entry-add-history 
              (string-append (car selection) " "))
          )
          ; write back the actual coordinate
          (gui-entry-add-history 
            (string-append (number->string (cadr selection)) " "
                           (number->string (caddr selection)) " "
                           (number->string (cadddr selection)) )
          )
        )
        ; its not valid point
        (begin
          ; if only one word is entered
          (if (null? (cdr selection))
            (let* 
              ( (words '()) 
                (keyword (car selection))
                (mesg "")
              )
              (for-each (lambda (y)
                  (if (and (<= (string-length keyword) (string-length y))
                           (string-ci=? keyword
                                        (substring y 0 (string-length keyword))) )
                    (set! words (cons y words))
                  )
                )
                *KEYWORDS_POINT_SELECTION*
              )
              ; determine what to do
              (case (length words)
                ; only one keyword is found
                ( (1)
                  (case (string->symbol (car words))
                    ; switch to viewing mode
                    ( (view)
                      (if (null? *SELECT_STACK*)
                        (begin
                          (selection-push-internal)
;                          (gui-update-gl-window #t)
                          (view-setup)
                        )
                        (gui-entry-add-history  "\nNo reentry to viewing ")
                      )
                    )
                  )
                )
                ; no internal keyword matches
                ( (0)
                  (if (null? *SELECT_KEYWORDS*)
                    (gui-entry-add-history 
                      (string-append "\n*Invalid entry* was " keyword))
                    ; check for externally defined keywords
                    (selection-keyword-internal keyword action)
                  )
                )
                ; several matching internal keywords
                (else
                  (set! mesg (string-append "\nAmbiguous keyword was " keyword " for ["))
                  (for-each (lambda (xx) 
                    (set! mesg (string-append mesg xx " ")) )
                    words
                  )
                  (set! mesg (string-append mesg "]"))
                  (gui-entry-add-history mesg)
                )
              )
            )
            (gui-entry-add-history " *Invalid entry*")
          )
        )
      )

      (if (and *SELECT_X* *SELECT_Y* *SELECT_Z*) ; all of them are set
        (begin
          (set! *SELECT_STATE* 'point)
          (action)
        )
        ; some are set, guide the user, when none set do not disturb
        (if (or *SELECT_X* *SELECT_Y* *SELECT_Z*)
          (let* ((tt " (need "))
            (if (not *SELECT_X*) (set! tt (string-append tt "X")))
            (if (not *SELECT_Y*) (set! tt (string-append tt "Y")))
            (if (not *SELECT_Z*) (set! tt (string-append tt "Z")))
            (set! tt (string-append *SELECT_TEXT_PREFIX* tt ")"))
            (gui-entry-label-set-text tt)
          )
        )
      )
    )
  )
)


; first stage function for distance selection, 
; it stores the starting point of a distance in *SELECT_BASE*
(define (selection-store-base)
  (if (eq? *SELECT_STATE* 'point)
    (begin
      (set! *SELECT_BASE* (list *SELECT_X* *SELECT_Y* *SELECT_Z*))
      (set! *SELECT_X* #f)
      (set! *SELECT_Y* #f)
      (set! *SELECT_Z* #f)
      ; change the text
      (set! *SELECT_TEXT_PREFIX* "Select second point:")
      (gui-entry-label-set-text *SELECT_TEXT_PREFIX*)
    )
  )
)

; second stage function for distance selection, 
; it calculates the distance between the two, specified points
(define (selection-calc-distance)
  (set! *SELECT_X* (- *SELECT_X* (car *SELECT_BASE*)))
  (set! *SELECT_Y* (- *SELECT_Y* (cadr *SELECT_BASE*)))
  (set! *SELECT_Z* (- *SELECT_Z* (caddr *SELECT_BASE*)))
)


; returns the distance in *SELECT_X* *SELECT_Y* *SELECT_Z*
; same internal keywords as for point selection
(define (selection-dist-point-internal selection action)
  (if (= (length selection) 0)
    (begin
      (if (member 'no_empty *SELECT_INIT*)
        (gui-entry-add-history " *Distance is required*")
        (begin
          (set! *SELECT_STATE* 'cancel)
          ; lets reset everything
          (action)
        )
      )
    )
    ; is it a distance definition and is it allowed to select it ?
    (if (and (string? (car selection))
             (string=? (car selection) "@")
             (not *SELECT_BASE*))
      (begin
        
        ; first, allow partial entry
        (case (length selection)
          ((2) (set! selection (append selection '(0 0))) )
          ((3) (set! selection (append selection '(0))) )
        )
        ; then validate
        (if (and (= (length selection) 4)
                 (number? (cadr selection))
                 (number? (caddr selection))
                 (number? (cadddr selection)) )
          (begin
            (set! *SELECT_X* (cadr selection))
            (set! *SELECT_Y* (caddr selection))
            (set! *SELECT_Z* (cadddr selection))
            (set! *SELECT_STATE* 'dist)
            (action)
          )
          (begin 
            (gui-entry-add-history " *Invalid distance*")
          )
        )
      )
      (begin
        ; normal point is supplied ?
        (if *SELECT_BASE*
          ; there is a base point, the second point is selected
          (begin
            (selection-coord-internal selection selection-calc-distance)
            (if (eq? *SELECT_STATE* 'point)
              (begin
                (set! *SELECT_STATE* 'dist)
                (action)
              )
              (set! *SELECT_STATE* 'init)
            )
          )
          ; this time the base point is specified
          (begin
            (selection-coord-internal selection selection-store-base)
            (set! *SELECT_STATE* 'init)
          )
        )
      )
    )
  )
)


;----------------------------------------------------------------------
; Public interfaces

(define (selection-object)
  (let*
    ( (tt (gui-entry-get-text))
      ; all number must be converted, integer and real
      ; this is necessary for centre selection, where coordinates are supplied
      ; however non-integer numbers are ignored for element selection
      (pp (map (lambda (x)
                 (if (number? (string->number x)) (string->number x) x))
               (text-split-with-empty tt #\, #f)
          )
      )
    )
    ; print out what the user entered
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text) " "))

    ; do the job
    (selection-object-internal pp *SELECT_ACTION*)
  )
)


(define (selection-coord)
  (let*
    ( (tt (gui-entry-get-text))
      (pp (map (lambda (x)
                 (if (string->number x) (string->number x) x))
               (text-split-with-empty tt #\, #f)
          )
      )
    )
    ; print out what the user entered
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text) " "))
    ; do the job
    (selection-coord-internal pp *SELECT_ACTION*)
  )
)


(define (selection-dist-point)
  (let*
    ( (tt (gui-entry-get-text))
      (pp (map (lambda (x)
                 (if (string->number x) (string->number x) x))
               (text-split-with-empty tt #\, #f)
          )
      )
    )
    ; if the first element has the @ sign, then create a new list where
    ; - the first element is the @ sign
    ; - the second element is the rest of string of the first element
    ;   converted into a number
    ; - the rest of the list is not changed
    (if (and (string? (car pp))
             (char=? #\@ (string-ref (car pp) 0)) )
      (set! pp (append (list "@" 
                             (if (string->number (substring (car pp) 1))
                               (string->number (substring (car pp) 1))
                               (substring (car pp) 1)
                             )
                       )
                       (cdr pp)
               )
      )
    )

    ; print out what the user entered
    (gui-entry-add-history (string-append "\n" (gui-entry-label-get-text) " "))
    ; do the job
    (selection-dist-point-internal pp *SELECT_ACTION*)
  )
)



(define (selection-cancel)
  (if (null? *DIALOG_WINDOWS*)
    (begin
      (set! *SELECT_STATE* 'cancel)
      ; notify user about cancellation
      (gui-entry-add-history "\n*Cancel*")
      ; reset all selection
      (selection-clear)

      ; update the screen, so selection is not shown
      (gui-update-gl-window #t)
      ; clear the command area
      (gui-entry-set-text "")
      #t
    )
    (begin
      (gui-entry-add-history "\nWARNING: close the dialog window(s)!")
      #f
    )
  )
)




