; OpenGL trackball implementation in scheme (Guile)
;--------------------------------------------------

; * (c) Copyright 1993, 1994, Silicon Graphics, Inc.
; * ALL RIGHTS RESERVED
; * Permission to use, copy, modify, and distribute this software for
; * any purpose and without fee is hereby granted, provided that the above
; * copyright notice appear in all copies and that both the copyright notice
; * and this permission notice appear in supporting documentation, and that
; * the name of Silicon Graphics, Inc. not be used in advertising
; * or publicity pertaining to distribution of the software without specific,
; * written prior permission.
; *
; * THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
; * AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
; * INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
; * FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL SILICON
; * GRAPHICS, INC.  BE LIABLE TO YOU OR ANYONE ELSE FOR ANY DIRECT,
; * SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY
; * KIND, OR ANY DAMAGES WHATSOEVER, INCLUDING WITHOUT LIMITATION,
; * LOSS OF PROFIT, LOSS OF USE, SAVINGS OR REVENUE, OR THE CLAIMS OF
; * THIRD PARTIES, WHETHER OR NOT SILICON GRAPHICS, INC.  HAS BEEN
; * ADVISED OF THE POSSIBILITY OF SUCH LOSS, HOWEVER CAUSED AND ON
; * ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
; * POSSESSION, USE OR PERFORMANCE OF THIS SOFTWARE.
; *
; * US Government Users Restricted Rights
; * Use, duplication, or disclosure by the Government is subject to
; * restrictions set forth in FAR 52.227.19(c)(2) or subparagraph
; * (c)(1)(ii) of the Rights in Technical Data and Computer Software
; * clause at DFARS 252.227-7013 and/or in similar or successor
; * clauses in the FAR or the DOD or NASA FAR Supplement.
; * Unpublished-- rights reserved under the copyright laws of the
; * United States.  Contractor/manufacturer is Silicon Graphics,
; * Inc., 2011 N.  Shoreline Blvd., Mountain View, CA 94039-7311.
; *
; * OpenGL(TM) is a trademark of Silicon Graphics, Inc.

(define-module (sx trackball))

(export trackball-calc-quaternion
        trackball-rotmatrix 
        trackball-add-quaternions
)

; counter to avoid error accumulation in the quaternions
(define *TRACKBALL_REFCOUNTER* 0)


; simulation of a trackball, points are projected onto a
; virtual trackball
; arguments must be between -1.0 and 1.0
(define (trackball-calc-quaternion p1x p1y p2x p2y)

  ; size of trackball
  ; it should be the distance between the rotation centre and
  ; the point on the object underneath the mouse
  (define TRACKBALLSIZE 0.2)

  ; project an x,y pair onto a sphere of radius r 
  ; or a hyperbolic sheet, if we are away from the center of the sphere
  (define (project_to_sphere r x y)
    (let*
      ((d (sqrt (+ (* x x) (* y y))))
       (t (/ r 1.41421356237309504880)) )
      (if (< d (* r 0.70710678118654752440))
        (sqrt (- (* r r) (* d d)))
        (/ (* t t) d)
      )
    )
  )

  ; cross product of two vectors 
  (define (vcross v1x v1y v1z v2x v2y v2z)
    (let*
      ((x (- (* v1y v2z) (* v1z v2y)))
       (y (- (* v1z v2x) (* v1x v2z)))
       (z (- (* v1x v2y) (* v1y v2x))) 
      )
      (list x y z)
    )
  )

  ; FIXME: check that the length is not zero !!!
  (define (vlength v)
    (let*
      ((x (car v))
       (y (cadr v))
       (z (caddr v))
      )
      (sqrt (+ (* x x) (* y y) (* z z)))
    )
  )

  ; compute quaternion
  (define (axis_to_quaternion a phi)
    (let*
      ((l (/ 1.0 (vlength a)))
       (sc (sin (/ phi 2.0)))
       (cc (cos (/ phi 2.0)))
      )
      (list (* (car a) l sc) (* (cadr a) l sc) (* (caddr a) l sc) cc)
    )
  )

  (if (and (= p1x p2x) (= p1y p2y))
    '(0.0 0.0 0.0 1.0) ; zero rotation
    (let*
      ((p1z (project_to_sphere TRACKBALLSIZE p1x p1y))
       (p2z (project_to_sphere TRACKBALLSIZE p2x p2y))
       (a   (vcross p2x p2y p2z p1x p1y p1z))
       (d   (list (- p1x p2x) (- p1y p2y) (- p1z p2z)))
       (t   (/ (vlength d) (* 2.0 TRACKBALLSIZE)))
       (t   (if (> t 1.0) 1.0 t))
       (t   (if (< t -1.0) -1.0 t))
       (phi (* 2.0 (asin t)))
      )
      ; given an axis and an angle, calculate quaternion
      (axis_to_quaternion a phi)
    )
  )
)


; build a rotation matrix from a quaternion
(define (trackball-rotmatrix q)
  (let*
    ((q0 (car q))
     (q1 (cadr q))
     (q2 (caddr q))
     (q3 (cadddr q))
     (m00 (- 1.0 (* 2.0 (+ (* q1 q1) (* q2 q2)))))
     (m01 (* 2.0 (- (* q0 q1) (* q2 q3))))
     (m02 (* 2.0 (+ (* q2 q0) (* q1 q3))))
     (m10 (* 2.0 (+ (* q0 q1) (* q2 q3))))
     (m11 (- 1.0 (* 2.0 (+ (* q2 q2) (* q0 q0)))))
     (m12 (* 2.0 (- (* q1 q2) (* q0 q3))))
     (m20 (* 2.0 (- (* q2 q0) (* q1 q3))))
     (m21 (* 2.0 (+ (* q1 q2) (* q0 q3))))
     (m22 (- 1.0 (* 2.0 (+ (* q1 q1) (* q0 q0)))))
    )
    (list m00 m01 m02 0.0 m10 m11 m12 0.0 m20 m21 m22 0.0 0.0 0.0 0.0 1.0)
  )
)


; given two rotations expressed as quaternions, calculate the
; equivalent single rotation and return it
(define (trackball-add-quaternions q1 q2)
  ; dot product of two vectors
  (define (vdot v1 v2)
    (+ (* (car v1) (car v2))
       (* (cadr v1) (cadr v2))
       (* (caddr v1) (caddr v2))
    )
  )

  (let*
    ((sq2 (cadddr q2))
     (sq1 (cadddr q1))
     ; t1 is q1 scaled by q2[3], it is directly used in tf
     ; t2 is q2 scaled by q1[3], it is directly used in tf
     ; t3 is the cross product of q1 and q2
     (t3x (- (* (cadr q2)  (caddr q1)) (* (caddr q2) (cadr q1))))
     (t3y (- (* (caddr q2) (car q1))   (* (car q2) (caddr q1))))
     (t3z (- (* (car q2)   (cadr q1))  (* (cadr q2) (car q1))))

     ; tf is a quaternion of t1+t2+t3 and the last component
     (tf (list (+ (* (car q1) sq2) (* (car q2) sq1) t3x)
               (+ (* (cadr q1) sq2) (* (cadr q2) sq1) t3y)
               (+ (* (caddr q1) sq2) (* (caddr q2) sq1) t3z)
               (- (* (cadddr q1) (cadddr q2)) (vdot q1 q2))))
    )

    ; increase the counter
    (set! *TRACKBALL_REFCOUNTER* (+ 1 *TRACKBALL_REFCOUNTER*))
    ; to avoid errors to creep in, normalise it after some time
    ; it is true for every quaternion that
    ; a^2 + b^2 + c^2 + d^2 = 1
    ; if they do not add up to one, then by dividing by the
    ; magnitude will renormalise the quaternion
    (if (> *TRACKBALL_REFCOUNTER* 100)
      (let*
        ((mag (+ (* (car tf) (car tf))
                 (* (cadr tf) (cadr tf))
                 (* (caddr tf) (caddr tf))
                 (* (cadddr tf) (cadddr tf))))
        )
        (set! *TRACKBALL_REFCOUNTER* 0)
        (list (/ (car tf) mag) (/ (cadr tf) mag)
              (/ (caddr tf) mag) (/ (cadddr tf) mag))
      )
      tf
    )
  )
)













